/* sqXWindow.c -- support for Unix and the X Window System.
 * 
 *   Copyright (C) 1996-2002 Ian Piumarta and other authors/contributors
 *     as listed elsewhere in this file.
 *   All rights reserved.
 *   
 *   This file is part of Unix Squeak.
 * 
 *   This file is distributed in the hope that it will be useful, but WITHOUT
 *   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *   FITNESS FOR A PARTICULAR PURPOSE.
 *   
 *   You may use and/or distribute this file ONLY as part of Squeak, under
 *   the terms of the Squeak License as described in `LICENSE' in the base of
 *   this distribution, subject to the following restrictions:
 * 
 *   1. The origin of this software must not be misrepresented; you must not
 *      claim that you wrote the original software.  If you use this software
 *      in a product, an acknowledgment to the original author(s) (and any
 *      other contributors mentioned herein) in the product documentation
 *      would be appreciated but is not required.
 * 
 *   2. This notice must not be removed or altered in any source distribution.
 * 
 *   Using (or modifying this file for use) in any context other than Squeak
 *   changes these copyright conditions.  Read the file `COPYING' in the
 *   directory `platforms/unix/doc' before proceeding with any such use.
 * 
 *   You are not allowed to distribute a modified version of this file
 *   under its original name without explicit permission to do so.  If
 *   you change it, rename it.
 */

/* Author: Ian Piumarta <ian.piumarta@inria.fr>
 *
 * Last edited: 2003-02-06 16:22:36 by piumarta on emilia.local.
 *
 * Support for displays deeper than 8 bits contributed by: Kazuki YASUMATSU
 *	<kyasu@crl.fujixerox.co.jp> <Kazuki.Yasumatsu@fujixerox.co.jp>
 *
 * Support for cursor and keypad editing keys based on code contributed by:
 *	Stefan Matthias Aust <sma@kiel.netsurf.de>
 *
 * Support for intelligent visual class selection contributed by:
 *	Bill Cattey <wdc@MIT.EDU>
 *
 * Support for European accented characters in selections, and
 * Support for displays shallower than 8 bits contributed by:
 *	Bert Freudenberg <bert@isg.cs.uni-magdeburg.de>
 *
 * Support for 24bpp TrueColour X display devices contributed by:
 *	Tim Rowledge <tim@sumeru.stanford.edu>
 *
 * Support for OSProcess plugin contributed by:
 *	Dave Lewis <lewis@mail.msen.com> Mon Oct 18 20:36:54 EDT 1999
 *
 * Support for browser plugins contributed by:
 *	Bert Freudenberg <bert@isg.cs.uni-magdeburg.de>
 *
 * BUGS: this file is too long; it should be split into two (Unix vs. X11).
 *	 icon stuff should be removed (window manager's responsibility).
 *	 RCS stuff has been removed.
 */

#include "sq.h"
extern "C" {
#include "aio.h"
}

#undef null
#undef DragEnter /* 1 */
#undef DragMove  /* 2 */
#undef DragLeave /* 3 */
#undef DragDrop  /* 4 */

#define NO_ICON
#define PRINT_PS_FORMS
#define SQ_FORM_FILENAME	"squeak-form.ppm"
#undef	FULL_UPDATE_ON_EXPOSE

#undef	DEBUG_EVENTS
#undef	DEBUG_SELECTIONS
#undef	DEBUG_BROWSER

/* if defined then the main Squeak window is a subwindow in an
   invisible enclosing InputOutput window.  helps integrate the
   Mozilla plugin stuff and fullscreen mode. */

#ifdef HAVE_LIBXEXT
# define USE_XSHM
#endif

/* OS_TYPE may be set in configure.in and passed via the Makefile */

#ifndef OS_TYPE
# ifdef UNIX
#   define OS_TYPE "unix"
# else
#  define OS_TYPE "unknown"
# endif
#endif

#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>

#ifdef HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif

#ifndef HEADLESS
# include <X11/Xlib.h>
# include <X11/Xutil.h>
# include <X11/Xatom.h>
# define XK_MISCELLANY
# define XK_XKB_KEYS
# include <X11/keysymdef.h>
# ifdef USE_XSHM
#   include <sys/ipc.h>
#   include <sys/shm.h>
#   include <X11/extensions/XShm.h>
# endif
# ifndef NO_ICON
#   include "squeakIcon.bitmap"
# endif
#endif

#if defined(HAVE_LIBDL)
# undef USE_INTERNAL_IMAGE
# include <dlfcn.h>
# if !defined(NDEBUG)
#   define NDEBUG
# endif
# include <assert.h>
# if defined(USE_INTERNAL_IMAGE)
#   include "zipio.h"
# endif
#endif

#include "sqQPEWindow.h"
#include <qpixmap.h>
#include <qimage.h>
#include <qpalette.h>
#include <qcolor.h>
#include <qfont.h>
#include <qdirectpainter_qws.h>
#include <qfileinfo.h>
#include <qmessagebox.h>
#ifdef NON_ZAURUS_ENVIRONMENT
#  include <qpainter.h>
#else 
#  include <qdirectpainter_qws.h>
#  ifdef USE_CFXGA
//#    include <custom.h>
#    include <fcntl.h>
#    include <sys/ioctl.h>
#    include "cfxga.h"
#    include <linux/fb.h>
#  endif
#endif
#include <qtoolbar.h>
#include <qlabel.h>

#define isAligned(T, V)	(((V) % sizeof(T)) == 0)
#define align(T, V)	(((V) / sizeof(T)) * sizeof(T))

/*** Variables -- Imported from Virtual Machine ***/
extern "C" {
  extern char *vmLibDir;
  extern unsigned char *memory;
  extern int interruptKeycode;
  extern int interruptPending;
  extern int interruptCheckCounter;
  extern int savedWindowSize;
  extern int fullScreenFlag;
  extern int specialObjectsOop;
  extern int nextWakeupTick;
  extern int printCallStack(void);
  extern int fetchIntegerofObject(int fieldIndex, int objectPointer);
  extern int fetchPointerofObject(int fieldIndex, int oop);
}

/*** Variables -- image and path names ***/
#define IMAGE_NAME_SIZE MAXPATHLEN

static char shortImageName[MAXPATHLEN+1];	/* image name */
//char        imageName[MAXPATHLEN+1];		/* full path to image */
static char vmPath[MAXPATHLEN+1];		/* full path to image directory */
static char vmName[MAXPATHLEN+1];		/* full path to vm */

#define DefaultHeapSize 16			/* megabytes */

int initialHeapSize= 0;

extern "C" {
  extern char *squeakPlugins;
}

/*** Variables -- globals for access from pluggable primitives ***/
int    argCnt= 0;
char **argVec= 0;
char **envVec= 0;

/* Pointers to arguments for VM.  Used by getAttributeIntoLength(). */
int    vmArgCnt= 0;
char **vmArgVec= 0;

/* Pointers to arguments for Squeak.  Used by getAttributeIntoLength(). */
int    squeakArgCnt= 0;
char **squeakArgVec= 0;

#if defined(USE_INTERNAL_IMAGE)
unsigned char *internalImage=    0;	/* non-zero means use internal image */
unsigned char *internalImagePtr= 0;	/* non-zero means plain image */
z_stream      *internalZStream=  0;	/* non-zero means gzipped image */
#endif

int		useItimer=	1;
unsigned int	lowResMSecs=	0;
#define	LOW_RES_TICK_MSECS	20	/* 1/50 second resolution */

char		*stPrimarySelection;	/* buffer holding selection */
char		*stEmptySelection= "";	/* immutable "empty string" value */
int		 stPrimarySelectionSize;/* size of buffer holding selection */
int		 stOwnsSelection= 0;	/* true if we own the X selection */
int		 stOwnsClipboard= 0;	/* true if we own the X clipboard */

int		 mapDelBs= 0;		/* 1 to map delete to backspace */
int		 swapBtn= 0;		/* 1 to swap yellow and blue buttons */
int		 optMapIndex= 0;	/* Option key modifier map index */
int		 cmdMapIndex= 0;	/* Command key modifier map index */

//unsigned int	 stColors[256];
int		 stRNMask, stGNMask, stBNMask;
int		 stRShift, stGShift, stBShift;
int		 headless= 0;

unsigned short	 stColors16[256];
#ifdef NON_ZAURUS_ENVIRONMENT
QImage		*stDisplayImage = NULL;
#else 
QDirectPainter *directp = NULL;
#endif

int qWidth = 0;
int qHeight = 0;
int qDepth = 0;

#ifdef IPAQ
bool portlait = true;
#else
bool portlait = false;
#endif

int skipShowDisplay = false;

SqueakWindow *squeak = NULL;
QPEApplication *stApp = NULL;

pthread_t interpThread = 0;
int interpreterStarted = false;

static FILE *logger = NULL;

#ifdef USE_CFXGA
static int cfxgaFd = 0;
/* unsigned char *projectorBuffer = NULL; */
static bool isCFXGAAvailable();
static bool CFXGAResumeCard();
static bool CFXGAIsActive();
static bool CFXGAActivate(bool needToClearScreen);
static bool CFXGAActivate2(bool needToClearScreen);
static void CFXGADeactivate();
static QSize CFXGAOutputSize();
static void CFXGASetDefaultSize();
/* static int CFXGAShowDisplay(unsigned char* vramPtr, int width, int height, int depth,
			    int affectedL, int affectedR, int affectedT, int affectedB);
*/
static int CFXGAShowDisplay2(unsigned char* vramPtr, int offset, int width, int height, int depth,
			    int affectedL, int affectedR, int affectedT, int affectedB);
static struct fb_fix_screeninfo fix;
#endif

void *startInterpreter(void *arg);
int setDisplayImageFunction(int depth, int bigEndianFlag);
void keyBufAppend(int keystate);
static int fillMemInfo();
unsigned int availableMemory();
Qt::WFlags defaultWFlags;
QPoint defaultOrigin = QPoint(0, 0);
QSizePolicy defaultSizePolicy;
QRect defaultGeometry;


/* events */
extern "C" {
  extern int inputEventSemaIndex;
}
#define IEB_SIZE	 64	/* must be power of 2 */

sqInputEvent inputEventBuffer[IEB_SIZE];
int iebIn=  0;	/* next IEB location to write */
int iebOut= 0;	/* next IEB location to read  */

#define SqueakWhite	0
#define SqueakBlack	1

int asmAlign= 1;
int sleepWhenUnmapped= 0;
int useJit= 0;
int jitProcs= 0;
int jitMaxPIC= 0;
int withSpy= 0;
int noTitle= 0;
int fullScreen= 0;
int iconified= 0;
int noEvents= 0;		/* 1 to disable new event handling */
int noSoundMixer= 0;		/* 1 to diable writing sound mixer levels */

struct timeval	 startUpTime;

int buttonState= 0;		/* mouse button state or 0 if not pressed */
int modifierState= 0;		/* modifier key state or 0 if none pressed */

#ifdef IPAQ
int nextMouseButtonPressed = Qt::NoButton;
#endif

/*** Functions ***/

#ifdef ioMSecs
# undef ioMSecs
#endif

static int  HandleEvents(void);
void RecordFullPathForVmName(char *localVmName);
void RecordFullPathForImageName(char *localImageName);
void SetUpTimers(void);
void usage(void);
void imageNotFound(char *imageName);
void ParseArguments(int argc, char **argv, int inHeader);
void sigsegv(int ignored);
void sighup(int ignored);

void SetColorEntry(int index, int red, int green, int blue);
void SetUpClipboard(void);
void SetUpPixmap(void);
void SetUpWindow(char *displayName);
void getMaskbit(unsigned long ul, int *nmask, int *shift);
void setupDownGradingColors(void);
void copyReverseImageBytes(int *fromImageData, int *toImageData,
			   int depth, int width, int height,
			   int affectedL, int affectedT, int affectedR, int affectedB);
void copyReverseImageWords(int *fromImageData, int *toImageData,
			   int depth, int width, int height,
			   int affectedL, int affectedT, int affectedR, int affectedB);

# define declareCopyFunction(NAME) \
  void NAME (int *fromImageData, int *toImageData, \
	     int width, int height, \
	     int affectedL, int affectedT, int affectedR, int affectedB)

declareCopyFunction(displayImage1To16);

declareCopyFunction(displayImage8To16);

declareCopyFunction(displayImage16To16);

declareCopyFunction(displayImage32To16);

# undef declareCopyFunction

typedef void (*displayImageFunc)(int *dispBitsIndex, int *destBitsIndex,
				 int width, int height,
				 int affectedL, int affectedT,
				 int affectedR, int affectedB);

displayImageFunc displayImage;

static void redrawDisplay(int l, int r, int t, int b);
static void redrawFullDisplay();

void    SetUpCharmap();
void    ux2st(unsigned char *string);
void    st2ux(unsigned char *string);
void    claimSelection(void);
void    sendSelection();
char   *getSelection(void);
static int  translateCode(QKeyEvent *qevt);
int    ioMSecs(void);
void   browserProcessCommand(void);       /* see sqUnixMozilla.c */

int    strtobkm(const char *str);

extern "C" {time_t convertToSqueakTime(time_t);}	/* unix epoch -> Squeak epoch */

static char modifierMap[16]= {
  //  0, 1, 2, 3, 4, 5, 6, 7,
  //0, 1, 2, 3, 8, 9, 10, 11,
  0, 1, 8, 9, 2, 3, 10, 11,
  0, 0, 0, 0, 0, 0,  0,  0
};

static char buttonMap[8]= {
  0, 4, 1, 5, 2, 6, 3, 7
};


/*** VM Home Directory Path ***/

int vmPathSize(void)
{
  return strlen(vmPath);
}

int vmPathGetLength(int sqVMPathIndex, int length)
{
  char *stVMPath= (char *)sqVMPathIndex;
  int count, i;

  count= strlen(vmPath);
  count= (length < count) ? length : count;

  /* copy the file name into the Squeak string */
  for (i= 0; i < count; i++)
    stVMPath[i]= vmPath[i];

  return count;
}

/* Conversion table from X to Squeak (reversible) */

unsigned char X_to_Squeak[256] =
{
  0,   1,   2,   3,   4,   5,   6,   7,       /*   0 -   7 */
  8,   9,   13,  11,  12,  10,  14,  15,      /*   8 -  15 */
  16,  17,  18,  19,  20,  21,  22,  23,      /*  16 -  23 */
  24,  25,  26,  27,  28,  29,  30,  31,      /*  24 -  31 */
  32,  33,  34,  35,  36,  37,  38,  39,      /*  32 -  39 */
  40,  41,  42,  43,  44,  45,  46,  47,      /*  40 -  47 */
  48,  49,  50,  51,  52,  53,  54,  55,      /*  48 -  55 */
  56,  57,  58,  59,  60,  61,  62,  63,      /*  56 -  63 */
  64,  65,  66,  67,  68,  69,  70,  71,      /*  64 -  71 */
  72,  73,  74,  75,  76,  77,  78,  79,      /*  72 -  79 */
  80,  81,  82,  83,  84,  85,  86,  87,      /*  80 -  87 */
  88,  89,  90,  91,  92,  93,  94,  95,      /*  88 -  95 */
  96,  97,  98,  99, 100, 101, 102, 103,      /*  96 - 103 */
  104, 105, 106, 107, 108, 109, 110, 111,     /* 104 - 111 */
  112, 113, 114, 115, 116, 117, 118, 119,     /* 112 - 119 */
  120, 121, 122, 123, 124, 125, 126, 127,     /* 120 - 127 */
  196, 197, 165, 201, 209, 247, 220, 225,     /* 128 - 135 */
  224, 226, 228, 227, 198, 176, 170, 248,     /* 136 - 143 */
  213, 206, 195, 207, 211, 212, 210, 219,     /* 144 - 151 */
  218, 221, 246, 245, 250, 249, 251, 252,     /* 152 - 159 */
  160, 193, 162, 163, 223, 180, 182, 164,     /* 160 - 167 */
  172, 169, 187, 199, 194, 173, 168, 255,     /* 168 - 175 */
  161, 177, 178, 179, 171, 181, 166, 183,     /* 176 - 183 */
  184, 185, 188, 200, 186, 189, 202, 192,     /* 184 - 191 */
  203, 231, 229, 204, 128, 129, 174, 130,     /* 192 - 199 */
  233, 131, 230, 232, 237, 234, 235, 236,     /* 200 - 207 */
  208, 132, 241, 238, 239, 205, 133, 215,     /* 208 - 215 */
  175, 244, 242, 243, 134, 217, 222, 167,     /* 216 - 223 */
  136, 135, 137, 139, 138, 140, 190, 141,     /* 224 - 231 */
  143, 142, 144, 145, 147, 146, 148, 149,     /* 232 - 239 */
  240, 150, 152, 151, 153, 155, 154, 214,     /* 240 - 247 */
  191, 157, 156, 158, 159, 253, 254, 216,     /* 248 - 255 */
};

unsigned char Squeak_to_X[256];

void SetUpCharmap()
{
  int i;
  for(i= 0; i < 256; i++)
    Squeak_to_X[X_to_Squeak[i]]= i;
}

void st2ux(unsigned char *string)
{
  if (!string) return;
  while (*string)
    {
     *string= Squeak_to_X[*string];
      string++;
    }
}

void ux2st(unsigned char *string)
{
  if (!string) return;
  while (*string)
    {
      *string= X_to_Squeak[*string];
      string++;
    }
}

/*** QT-related Functions ***/

void
SqueakWindow::SetupQPEDisplay()
{
  SetUpCharmap();
  SetUpClipboard();
  defaultWFlags = getWFlags();
  defaultOrigin.setX(geometry().x());
  defaultOrigin.setY(geometry().y());
  defaultSizePolicy = sizePolicy();
  defaultGeometry = geometry();
#ifdef USE_CFXGA
  CFXGAActivate2(true);
#endif /*  USE_CFXGA */
  /*SetUpWindow(); */
  SetUpPixmap();
}

void closeQTDisplay()
{
#ifdef NON_ZAURUS_ENVIRONMENT
  if (stDisplayImage != NULL) {
    delete stDisplayImage;
  }
  stDisplayImage = NULL;
#endif
}

static void noteResize(int w, int h)
{
}


static int resized(void)
{
  return 0;
}

void claimSelection(void)
{
}

void sendSelection()
{
}

char *getSelection(void)
{
  return "";
}

/* a modified copy of fullDisplayUpdate().
   Note: if the format of Form or Bitmap changes, or if
   the special object index of Display is changed, this
   version of the code WILL FAIL!  Otherwise it is to be
   preferred.
*/
#define longAt(i) (*((int *) (i)))
static void redrawFullDisplay()
{
  int displayObj;
  int w;
  int h;
  int d;
  int bitsIndex;

  displayObj = longAt(((((char *) specialObjectsOop)) + 4) + (14 << 2));
  bitsIndex = fetchPointerofObject(0, displayObj) + 4;
  w = fetchIntegerofObject(1, displayObj);
  h = fetchIntegerofObject(2, displayObj);
  d = fetchIntegerofObject(3, displayObj);
  ioShowDisplay(bitsIndex, w, h, d, 0, w, 0, h);
}

static void redrawDisplay(int aX, int aY, int aW, int aH)
{
  int displayObj;
  int w;
  int h;
  int d;
  int bitsIndex;

  displayObj = longAt(((((char *) specialObjectsOop)) + 4) + (14 << 2));
  bitsIndex = fetchPointerofObject(0, displayObj) + 4;
  w = fetchIntegerofObject(1, displayObj);
  h = fetchIntegerofObject(2, displayObj);
  d = fetchIntegerofObject(3, displayObj);
  ioShowDisplay(bitsIndex, w, h, d, aX, aY, aW, aH);
}

#undef longAt


int mousePositionX = 0;
int mousePositionY = 0;

/*** event handling ***/


/* set asynchronous input event semaphore
 */
int ioSetInputSemaphore(int semaIndex)
{
  if ((semaIndex == 0) || (noEvents == 1))
    {
      success(false);
    }
  else
    {
      inputEventSemaIndex= semaIndex;
    }
  return true;
}


#define iebEmptyP()	(iebIn == iebOut)
#define iebAdvance(P)	(P= ((P + 1) & (IEB_SIZE - 1)))


/* retrieve the next input event from the OS
 */
int ioGetNextEvent(sqInputEvent *evt)
{
  if (iebEmptyP()) ioProcessEvents();
  if (iebEmptyP()) return false;
  *evt= inputEventBuffer[iebOut];
  iebAdvance(iebOut);
  return true;
}


static sqInputEvent *allocateInputEvent(int eventType)
{
  sqInputEvent *evt= &inputEventBuffer[iebIn];
  iebAdvance(iebIn);
  if (iebEmptyP())
    {
      /* overrun: discard oldest event */
      iebAdvance(iebOut);
    }
  evt->type= eventType;
  evt->timeStamp= ioMSecs();
  return evt;
}

#define allocateMouseEvent() ( \
  (sqMouseEvent *)allocateInputEvent(EventTypeMouse) \
)

#define allocateKeyboardEvent() ( \
  (sqKeyboardEvent *)allocateInputEvent(EventTypeKeyboard) \
)

static void signalInputEvent(void)
{
#ifdef DEBUG_EVENTS
  printf("signalInputEvent\n");
#endif
  if (inputEventSemaIndex > 0)
    signalSemaphoreWithIndex(inputEventSemaIndex);
}

#ifdef DEBUG_EVENTS
#include <ctype.h>
static void printKey(int key)
{
  printf(" `%c' (%d = 0x%x)", (isgraph(key) ? key : ' '), key, key);
}
#endif

static int x2sqButton(int button)
{
  /* ASSUME: (button >= 1) & (button <= 3) */
  static int rybMap[4]= { 0, RedButtonBit, YellowButtonBit, BlueButtonBit };
  static int rbyMap[4]= { 0, RedButtonBit, BlueButtonBit, YellowButtonBit };
  return (swapBtn ? rbyMap : rybMap)[button];
}

#ifdef DEBUG_EVENTS
static void printButtons(int buttons)
{
  if (buttons & RedButtonBit)    printf(" red");
  if (buttons & YellowButtonBit) printf(" yellow");
  if (buttons & BlueButtonBit)   printf(" blue");
}
#endif

static int qt2sqModifier(int state)
{
  return 0;
}

#ifdef DEBUG_EVENTS
static void printModifiers(int midofiers)
{
  if (midofiers & ShiftKeyBit)   printf(" Shift");
  if (midofiers & CtrlKeyBit)    printf(" Control");
  if (midofiers & CommandKeyBit) printf(" Command");
  if (midofiers & OptionKeyBit)  printf(" Option");
}
#endif


static int getButtonState(void)
{
  /* red button honours the modifiers:
   *	red+ctrl    = yellow button
   *	red+command = blue button
   */
  int buttons= buttonState;
  int modifiers= modifierState;
  if ((buttons == RedButtonBit) && modifiers)
    {
      int yellow= swapBtn ? BlueButtonBit   : YellowButtonBit;
      int blue=   swapBtn ? YellowButtonBit : BlueButtonBit;
      switch (modifiers)
	{
	case CtrlKeyBit:    buttons= yellow; modifiers &= ~CtrlKeyBit;    break;
	case CommandKeyBit: buttons= blue;   modifiers &= ~CommandKeyBit; break;
	}
    }
#ifdef DEBUG_EVENTS_DISABLED
  printf("BUTTONS");
  printModifiers(modifiers);
  printButtons(buttons);
  printf("\n");
#endif
  return buttons | (modifiers << 3);
}


static void
recordMouseEvent(QMouseEvent *qevt)
{
  unsigned int state = qevt->state();

#ifdef IPAQ
  if (nextMouseButtonPressed == Qt::NoButton)
    buttonState = buttonMap[state & QMouseEvent::MouseButtonMask];
  else
    {
    buttonState = buttonMap[nextMouseButtonPressed];
    nextMouseButtonPressed = Qt::NoButton;
    }
#else
  buttonState = buttonMap[state & QMouseEvent::MouseButtonMask];
#endif

#ifdef IPAQ
  if (portlait) {
    mousePositionX = qevt->x();
    mousePositionY = qevt->y();
  } else {
    mousePositionX = qWidth - 1 - qevt->y();
    mousePositionY = qevt->x();
  }
#else
  mousePositionX = qevt->x();
  mousePositionY = qevt->y();
#endif

  if (inputEventSemaIndex == 0) {
    return;
  }
  
  sqMouseEvent *evt= allocateMouseEvent();
  evt->x = mousePositionX;
  evt->y = mousePositionY;

  if ((modifierState & CommandKeyBit) && (buttonState == 4)) {
    buttonState = 1;
  }
  if ((modifierState & ShiftKeyBit)) {
    buttonState = 2;
  }
  
  evt->buttons = buttonState;
  evt->modifiers = modifierState;
  evt->reserved1= 0;
  evt->reserved2= 0;
}

static void
recordKeyboardEvent(QKeyEvent *qevt, int pressCode)
{
#ifdef IPAQ
  int keyCode = qevt->key();
  if ((keyCode == 4152) // IPAQ device button 1 (calender)
    || (keyCode == 4156) // IPAQ device button 3 (mail)
    || (keyCode == 4167)) // IPAQ device button 5 (voice recorder)
    {
    if (nextMouseButtonPressed == Qt::RightButton)
      nextMouseButtonPressed = Qt::NoButton;
    else
      nextMouseButtonPressed = Qt::RightButton;
    return;
    }
  if ((keyCode == 4153) // IPAQ device button 2 (contacts)
    || (keyCode == 4155)) // IPAQ device button 4 (apps)
    {
    if (nextMouseButtonPressed == Qt::MidButton)
      nextMouseButtonPressed = Qt::NoButton;
    else
      nextMouseButtonPressed = Qt::MidButton;
    return;
    }
  return;
#else
  int charCode= 0;
  int keystate= 0;
  charCode = translateCode(qevt);
  if (charCode == 0) {
    return;
  }

  if (charCode > 255) {
    if (pressCode == EventKeyDown) {
      modifierState |= modifierMap[(charCode>>8) & 0xF]; /* |= */
    } else if (pressCode == EventKeyUp) {
      modifierState = 0; //&= (~(modifierMap[(charCode>>8) & 0xF]) & 0xF);
    }
    if ((charCode & 0xFF) == 0) {
      return;
    } else {
      charCode = charCode & 0xFF;
    }
  } else {
    modifierState = 0;
    charCode = X_to_Squeak[charCode];
  }
  keystate= charCode | (modifierState << 8);
  if (logger != NULL) {
    fprintf(logger, "keystate: %x\n", keystate);
  }
  if (keystate == interruptKeycode) {
    /* Note: interrupt key is "meta"; it not reported as a keystroke */
    interruptPending= true;
    interruptCheckCounter= 0;
  } else {
    if (pressCode == EventKeyDown) {
      keyBufAppend(keystate);
    }
    if (inputEventSemaIndex != 0) {
      sqKeyboardEvent *evt= allocateKeyboardEvent();
      evt->charCode= charCode;
      evt->pressCode= pressCode;
      evt->modifiers= modifierState;
      evt->reserved1=
	evt->reserved2=
	evt->reserved3= 0;
    }
  }
#endif /* else IPAQ */
}

void
SqueakWindow::resizeEvent(QResizeEvent * e)
{
  QMainWindow::resizeEvent(e);

  portlait = (width() <= height());

  SetUpWindow();
}

void
SqueakWindow::mousePressEvent(QMouseEvent *e)
{
  recordMouseEvent(e);
}

void
SqueakWindow::mouseReleaseEvent(QMouseEvent *e)
{
  recordMouseEvent(e);
}

void
SqueakWindow::mouseMoveEvent(QMouseEvent *e)
{
  recordMouseEvent(e);
}

void
SqueakWindow::mouseDoubleClickEvent(QMouseEvent *e)
{
  recordMouseEvent(e);
}

void
SqueakWindow::keyPressEvent(QKeyEvent *e)
{
  recordKeyboardEvent(e, EventKeyDown);
  recordKeyboardEvent(e, EventKeyChar);
}

void
SqueakWindow::keyReleaseEvent(QKeyEvent *e)
{
  recordKeyboardEvent(e, EventKeyUp);
}

void
SqueakWindow::closeEvent(QCloseEvent *ce)
{
  skipShowDisplay = true;
  switch (QMessageBox::information(this, "Squeak Dialog",
				   "Quit Squeak without saving?",
				   "Yes", "No", 0, 1)) {
  case 0:
    ce->accept();
    break;
  case 1:
  default:
    ce->ignore();
    break;
  } 
  skipShowDisplay = false;
}

/*
void
SqueakWindow::paintEvent(QPaintEvent *e)
{
  QRect rect = e->rect();

  if (!interpreterStarted) {
    QMainWindow::paintEvent(e);
    return;
  }
  redrawDisplay(rect.x(), rect.y(), rect.width(), rect.height());
}
*/


int HandleEvents(void)
{
  return 0;
}


void getMaskbit(unsigned long ul, int *nmask, int *shift)
{
  int i;
  unsigned long hb;

  *nmask= *shift= 0;
  hb= 0x8000;  hb= (hb<<16);  /* hb = 0x80000000UL */
  for (i= 31; ((ul & hb) == 0) && i >= 0; --i, ul<<= 1)
    ;
  for (; ((ul & hb) != 0) && i >= 0; --i, ul<<= 1)
    (*nmask)++;
  *shift= i+1;
}

void SetColorEntry(int index, int red, int green, int blue)
{
  unsigned int r, g, b;

  if (index >= 256)
    return;

  r= red;
  g= green;
  b= blue;

  stColors16[index]= ((((r>>(16-stRNMask)) & 0x1F) << stRShift) |
		      (((g>>(16-stGNMask)) & 0x3F) << stGShift) |
		      (((b>>(16-stBNMask)) & 0x1F) << stBShift));
}

void SetUpPixmap(void)
{
  stRNMask = 5;
  stRShift = 11;
  stGNMask = 6;
  stGShift = 5;
  stBNMask = 5;
  stBShift = 0;

  /* 1-bit colors (monochrome) */
  SetColorEntry(0, 65535, 65535, 65535);	/* white or transparent */
  SetColorEntry(1,     0,     0,     0);	/* black */

	/* additional colors for 2-bit color */
  SetColorEntry(2, 65535, 65535, 65535);	/* opaque white */
  SetColorEntry(3, 32768, 32768, 32768);	/* 1/2 gray */

	/* additional colors for 4-bit color */
  SetColorEntry( 4, 65535,     0,     0);	/* red */
  SetColorEntry( 5,     0, 65535,     0);	/* green */
  SetColorEntry( 6,     0,     0, 65535);	/* blue */
  SetColorEntry( 7,     0, 65535, 65535);	/* cyan */
  SetColorEntry( 8, 65535, 65535,     0);	/* yellow */
  SetColorEntry( 9, 65535,     0, 65535);	/* magenta */
  SetColorEntry(10,  8192,  8192,  8192);	/* 1/8 gray */
  SetColorEntry(11, 16384, 16384, 16384);	/* 2/8 gray */
  SetColorEntry(12, 24576, 24576, 24576);	/* 3/8 gray */
  SetColorEntry(13, 40959, 40959, 40959);	/* 5/8 gray */
  SetColorEntry(14, 49151, 49151, 49151);	/* 6/8 gray */
  SetColorEntry(15, 57343, 57343, 57343);	/* 7/8 gray */

	/* additional colors for 8-bit color */
	/* 24 more shades of gray (does not repeat 1/8th increments) */
  SetColorEntry(16,  2048,  2048,  2048);	/*  1/32 gray */
  SetColorEntry(17,  4096,  4096,  4096);	/*  2/32 gray */
  SetColorEntry(18,  6144,  6144,  6144);	/*  3/32 gray */
  SetColorEntry(19, 10240, 10240, 10240);	/*  5/32 gray */
  SetColorEntry(20, 12288, 12288, 12288);	/*  6/32 gray */
  SetColorEntry(21, 14336, 14336, 14336);	/*  7/32 gray */
  SetColorEntry(22, 18432, 18432, 18432);	/*  9/32 gray */
  SetColorEntry(23, 20480, 20480, 20480);	/* 10/32 gray */
  SetColorEntry(24, 22528, 22528, 22528);	/* 11/32 gray */
  SetColorEntry(25, 26624, 26624, 26624);	/* 13/32 gray */
  SetColorEntry(26, 28672, 28672, 28672);	/* 14/32 gray */
  SetColorEntry(27, 30720, 30720, 30720);	/* 15/32 gray */
  SetColorEntry(28, 34815, 34815, 34815);	/* 17/32 gray */
  SetColorEntry(29, 36863, 36863, 36863);	/* 18/32 gray */
  SetColorEntry(30, 38911, 38911, 38911);	/* 19/32 gray */
  SetColorEntry(31, 43007, 43007, 43007);	/* 21/32 gray */
  SetColorEntry(32, 45055, 45055, 45055);	/* 22/32 gray */
  SetColorEntry(33, 47103, 47103, 47103);	/* 23/32 gray */
  SetColorEntry(34, 51199, 51199, 51199);	/* 25/32 gray */
  SetColorEntry(35, 53247, 53247, 53247);	/* 26/32 gray */
  SetColorEntry(36, 55295, 55295, 55295);	/* 27/32 gray */
  SetColorEntry(37, 59391, 59391, 59391);	/* 29/32 gray */
  SetColorEntry(38, 61439, 61439, 61439);	/* 30/32 gray */
  SetColorEntry(39, 63487, 63487, 63487);	/* 31/32 gray */

  /* The remainder of color table defines a color cube with six steps
     for each primary color. Note that the corners of this cube repeat
     previous colors, but simplifies the mapping between RGB colors and
     color map indices. This color cube spans indices 40 through 255.
     */
  {
    int r, g, b;

    for (r= 0; r < 6; r++)
      for (g= 0; g < 6; g++)
	for (b= 0; b < 6; b++)
	  {
	    int i= 40 + ((36 * r) + (6 * b) + g);
	    if (i > 255) error("index out of range in color table compuation");
	    SetColorEntry(i, (r * 65535) / 5, (g * 65535) / 5, (b * 65535) / 5);
	  }
  }

}


SqueakWindow::SqueakWindow(QWidget* parent, const char* name, WFlags f)
  : QMainWindow(parent, name, f)
{
  //resize(232, 258);
  fileSelector = NULL;

  //logger = fopen("/tmp/squeak.log", "w+");
  
  /* fprintf(stderr, "width = %d, height = %d\n", this->width(), this->height()); */
}

SqueakWindow::~SqueakWindow()
{
  if (logger != NULL) {
    fclose(logger);
    logger = NULL;
  }
#ifdef USE_CFXGA
  if (cfxgaFd != 0) {
    CFXGADeactivate();
  }
#endif
}

void
SqueakWindow::SetUpWindow()
{
  static int count = 0;

  if (count > 0) {
    return;
  }

  count++;
  
  if (fullScreenFlag || fullScreen) {
    if (size() !=  stApp->desktop()->size()) {
      enableFullScreen();
    }
  } else {
    if (count == 0) {
      disableFullScreen();
    }
  }
  
#ifdef IPAQ
  if (portlait) {
    qWidth = this->width();
    qHeight = this->height();
  } else {
    qWidth = this->height();
    qHeight = this->width();
    }
#else
  qWidth = this->width();
  qHeight = this->height();
#endif
#ifdef NON_ZAURUS_ENVIRONMENT
  if (stDisplayImage != NULL) {
    delete stDisplayImage;
  }
  stDisplayImage = new QImage(qWidth, qHeight, 16);
#endif
  if (qDepth == 0) {
    setDisplayImageFunction(16, true);
  } else {
    setDisplayImageFunction(qDepth, true);
  }
  count--;
}

/*** Event Recording Functions ***/


static int translateCode(QKeyEvent *qevt)
{
  int key = qevt->key();
  int mod = ((qevt->state()) >> 3) & 0xF;

  if (logger != NULL) {
    fprintf(logger, "mod(1): %x\n", mod);
    fprintf(logger, "key: %x\n", key);
  }

  switch (key) {
  case Qt::Key_Space:
  case Qt::Key_Exclam:
  case Qt::Key_QuoteDbl:
  case Qt::Key_NumberSign:
  case Qt::Key_Dollar:
  case Qt::Key_Percent:
  case Qt::Key_Ampersand:
  case Qt::Key_Apostrophe:
  case Qt::Key_ParenLeft:
  case Qt::Key_ParenRight:
  case Qt::Key_Asterisk:
  case Qt::Key_Plus:
  case Qt::Key_Comma:
  case Qt::Key_Minus:
  case Qt::Key_Period:
  case Qt::Key_Slash:
  case Qt::Key_0:
  case Qt::Key_1:
  case Qt::Key_2:
  case Qt::Key_3:
  case Qt::Key_4:
  case Qt::Key_5:
  case Qt::Key_6:
  case Qt::Key_7:
  case Qt::Key_8:
  case Qt::Key_9:
  case Qt::Key_Colon:
  case Qt::Key_Semicolon:
  case Qt::Key_Less:
  case Qt::Key_Equal:
  case Qt::Key_Greater:
  case Qt::Key_Question:
  case Qt::Key_At:
  case Qt::Key_A:
  case Qt::Key_B:
    //case Qt::Key_C:
  case Qt::Key_D:
  case Qt::Key_E:
  case Qt::Key_F:
  case Qt::Key_G:
  case Qt::Key_H:
  case Qt::Key_I:
  case Qt::Key_J:
  case Qt::Key_K:
  case Qt::Key_L:
  case Qt::Key_M:
  case Qt::Key_N:
  case Qt::Key_O:
  case Qt::Key_P:
  case Qt::Key_Q:
  case Qt::Key_R:
  case Qt::Key_S:
  case Qt::Key_T:
  case Qt::Key_U:
    //case Qt::Key_V:
  case Qt::Key_W:
    //case Qt::Key_X:
  case Qt::Key_Y:
    //case Qt::Key_Z:
  case Qt::Key_BracketLeft:
  case Qt::Key_Backslash:
  case Qt::Key_BracketRight:
  case Qt::Key_AsciiCircum:
  case Qt::Key_Underscore:
  case Qt::Key_QuoteLeft:
  case Qt::Key_BraceLeft:
  case Qt::Key_Bar:
  case Qt::Key_BraceRight:
  case Qt::Key_AsciiTilde:
  case Qt::Key_nobreakspace:
  case Qt::Key_exclamdown:
  case Qt::Key_cent:
  case Qt::Key_sterling:
  case Qt::Key_currency:
  case Qt::Key_yen:
  case Qt::Key_brokenbar:
  case Qt::Key_section:
  case Qt::Key_diaeresis:
  case Qt::Key_copyright:
  case Qt::Key_ordfeminine:
  case Qt::Key_notsign:
  case Qt::Key_hyphen:
  case Qt::Key_registered:
  case Qt::Key_macron:
  case Qt::Key_degree:
  case Qt::Key_plusminus:
  case Qt::Key_twosuperior:
  case Qt::Key_threesuperior:
  case Qt::Key_acute:
  case Qt::Key_mu:
  case Qt::Key_paragraph:
  case Qt::Key_periodcentered:
  case Qt::Key_cedilla:
  case Qt::Key_onesuperior:
  case Qt::Key_masculine:
  case Qt::Key_guillemotright:
  case Qt::Key_onequarter:
  case Qt::Key_onehalf:
  case Qt::Key_threequarters:
  case Qt::Key_questiondown:
  case Qt::Key_Agrave:
  case Qt::Key_Aacute:
  case Qt::Key_Acircumflex:
  case Qt::Key_Atilde:
  case Qt::Key_Adiaeresis:
  case Qt::Key_Aring:
  case Qt::Key_AE:
  case Qt::Key_Ccedilla:
  case Qt::Key_Egrave:
  case Qt::Key_Eacute:
  case Qt::Key_Ecircumflex:
  case Qt::Key_Ediaeresis:
  case Qt::Key_Igrave:
  case Qt::Key_Iacute:
  case Qt::Key_Icircumflex:
  case Qt::Key_Idiaeresis:
  case Qt::Key_ETH:
  case Qt::Key_Ntilde:
  case Qt::Key_Ograve:
  case Qt::Key_Oacute:
  case Qt::Key_Ocircumflex:
  case Qt::Key_Otilde:
  case Qt::Key_Odiaeresis:
  case Qt::Key_multiply:
  case Qt::Key_Ooblique:
  case Qt::Key_Ugrave:
  case Qt::Key_Uacute:
  case Qt::Key_Ucircumflex:
  case Qt::Key_Udiaeresis:
  case Qt::Key_Yacute:
  case Qt::Key_THORN:
  case Qt::Key_ssharp:
  case Qt::Key_agrave:
  case Qt::Key_aacute:
  case Qt::Key_acircumflex:
  case Qt::Key_atilde:
  case Qt::Key_adiaeresis:
  case Qt::Key_aring:
  case Qt::Key_ae:
  case Qt::Key_ccedilla:
  case Qt::Key_egrave:
  case Qt::Key_eacute:
  case Qt::Key_ecircumflex:
  case Qt::Key_ediaeresis:
  case Qt::Key_igrave:
  case Qt::Key_iacute:
  case Qt::Key_icircumflex:
  case Qt::Key_idiaeresis:
  case Qt::Key_eth:
  case Qt::Key_ntilde:
  case Qt::Key_ograve:
  case Qt::Key_oacute:
  case Qt::Key_ocircumflex:
  case Qt::Key_otilde:
  case Qt::Key_odiaeresis:
  case Qt::Key_division:
  case Qt::Key_oslash:
  case Qt::Key_ugrave:
  case Qt::Key_uacute:
  case Qt::Key_ucircumflex:
  case Qt::Key_udiaeresis:
  case Qt::Key_yacute:
  case Qt::Key_thorn:
  case Qt::Key_ydiaeresis:
    return qevt->ascii() | (mod<<8);
  case Qt::Key_Z:
  case Qt::Key_X:
  case Qt::Key_C:
  case Qt::Key_V:
    {
      int k;
      if (mod & 1) {
	k = qevt->key();
      } else {
	if (mod == 0) {
	  k = qevt->ascii();
	} else {
	  k = qevt->key() + 0x20;
	}
      }
      return k | (mod<<8);
    }
  case Qt::Key_F22:
    return (2<<8);
  case Qt::Key_F11:
    return (2<<8) | '.';
  case Qt::Key_Escape:
    return 3 | (mod<<8);
  case Qt::Key_Tab:
    return 9 | (mod<<8);
  case Qt::Key_Backtab:
    return 0;
  case Qt::Key_Backspace:
    return 8 | (mod<<8);
  case Qt::Key_Return:
  case Qt::Key_Enter:
    return 10 | (mod<<8);
    //case 3:
    //return 'c' | (2<<8);
  case 0x2014:
    return 'q' | (2<<8);
  case 0x2015:
    return 'o' | (2<<8);
  case 0x2016:
    return 'p' | (2<<8);
  case 0x2017:
    return 'a' | (2<<8);
  case 0x2018:
    return 's' | (2<<8);
  case 0x2019:
    return 'n' | (2<<8);
  case 0x201a:
    return 'm' | (2<<8);
  case 22:
    return 'v' | (2<<8);
  case 24:
    return 'x' | (2<<8);
  case Qt::Key_Insert:
  case Qt::Key_Delete:
  case Qt::Key_Pause:
  case Qt::Key_Print:
  case Qt::Key_SysReq:
  case Qt::Key_Home:
  case Qt::Key_End:
    return 0;
  case Qt::Key_Left:
    return 28;
  case Qt::Key_Up:
    return 30;
  case Qt::Key_Right:
    return 29;
  case Qt::Key_Down:
    return 31;
  case Qt::Key_Prior: /* = Qt::Key_PageUp */
    return 11; /* page up */
  case Qt::Key_Next:  /* = Qt::Key_PageDown */
    return 12; /* page down */
  case Qt::Key_Shift:
    return 0 | (1<<8);
  case Qt::Key_Control:
    return 0 | (4<<8);
  case Qt::Key_Meta:
  case Qt::Key_Alt:
    return 0 | (2<<8);
  case Qt::Key_CapsLock:
  case Qt::Key_NumLock:
  case Qt::Key_ScrollLock:
  case Qt::Key_F1:
  case Qt::Key_F2:
  case Qt::Key_F3:
  case Qt::Key_F4:
  case Qt::Key_F5:
  case Qt::Key_F6:
  case Qt::Key_F7:
  case Qt::Key_F8:
  case Qt::Key_F9:
  case Qt::Key_F10:
    //case Qt::Key_F11:
  case Qt::Key_F12:
  case Qt::Key_F13:
  case Qt::Key_F14:
  case Qt::Key_F15:
  case Qt::Key_F16:
  case Qt::Key_F17:
  case Qt::Key_F18:
  case Qt::Key_F19:
  case Qt::Key_F20:
  case Qt::Key_F21:
    //case Qt::Key_F22:
  case Qt::Key_F23:
  case Qt::Key_F24:
  case Qt::Key_F25:
  case Qt::Key_F26:
  case Qt::Key_F27:
  case Qt::Key_F28:
  case Qt::Key_F29:
  case Qt::Key_F30:
  case Qt::Key_F31:
  case Qt::Key_F32:
  case Qt::Key_F33:
  case Qt::Key_F34:
  case Qt::Key_F35:
  case Qt::Key_Menu:
  case Qt::Key_Help:
    return 0;
  default:
    return 0;
  }
  return 0;
}


/*** the following is deprecated and should really go away.  for now we keep it
     for backwards compatibility with ancient images ***/

#define KEYBUF_SIZE 64

static int keyBuf[KEYBUF_SIZE];		/* circular buffer */
static int keyBufGet= 0;		/* index of next item of keyBuf to read */
static int keyBufPut= 0;		/* index of next item of keyBuf to write */
static int keyBufOverflows= 0;		/* number of characters dropped */

int ioPeekKeystroke(void)			/* DEPRECATED */
{
  int keystate;

  ioProcessEvents();  /* process all pending events */

  if (keyBufGet == keyBufPut)
    return -1;  /* keystroke buffer is empty */

  keystate= keyBuf[keyBufGet];
  /* set modifer bits in buttonState to reflect the last keystroke peeked at */
  /* buttonState= ((keystate >> 5) & 0xF8) | (buttonState & 0x7); */
  return keystate;
}


int ioGetKeystroke(void)			/* DEPRECATED */
{
  int keystate;

  ioProcessEvents();  /* process all pending events */
  if (keyBufGet == keyBufPut)
    return -1;  /* keystroke buffer is empty */

  keystate= keyBuf[keyBufGet];
  keyBufGet= (keyBufGet + 1) % KEYBUF_SIZE;
  /* set modifer bits in buttonState to reflect the last keystroke fetched */
  /* buttonState= ((keystate >> 5) & 0xF8) | (buttonState & 0x7); */

  return keystate;
}

void keyBufAppend(int keystate)
{
  /* bug: this should be rewritten to cope with nConv > 1 */
  keyBuf[keyBufPut]= keystate;
  keyBufPut= (keyBufPut + 1) % KEYBUF_SIZE;
  if (keyBufGet == keyBufPut)
    {
      /* buffer overflow; drop the last character */
      keyBufGet= (keyBufGet + 1) % KEYBUF_SIZE;
      keyBufOverflows++;
    }
}


/*** I/O Primitives ***/


#ifndef HAVE_SNPRINTF
# ifdef HAVE___SNPRINTF	/* Solaris 2.5 */
    extern int __snprintf(char *buf, size_t limit, const char *fmt, ...);
#   define snprintf __snprintf
#   define HAVE_SNPRINTF
# endif
#endif


int ioFormPrint(int bitsAddr, int width, int height, int depth,
		double hScale, double vScale, int landscapeFlag)
{
  return false;
}


int ioBeep(void)
{
  QPEApplication::beep();
  return true;
}


int ioGetButtonState(void)
{
  ioProcessEvents();  /* process all pending events */
  return getButtonState();
}


int ioLowResMSecs(void)
{
  return (useItimer)
    ? lowResMSecs
    : ioMSecs();
}


int ioMSecs(void)
{
  struct timeval now;
  gettimeofday(&now, 0);
  if ((now.tv_usec-= startUpTime.tv_usec) < 0)
    {
      now.tv_usec+= 1000000;
      now.tv_sec-= 1;
    }
  now.tv_sec-= startUpTime.tv_sec;
  return (now.tv_usec / 1000 + now.tv_sec * 1000);
}


int ioMicroMSecs(void)
{
  /* return the highest available resolution of the millisecond clock */
  return ioMSecs();	/* this already to the nearest millisecond */
}


int ioRelinquishProcessorForMicroseconds(int microSeconds)
{
  if (HandleEvents())
    {
      aioPoll(0);
      return 0;
    }
  else
    aioPoll(microSeconds);
  return 0;
}


int ioProcessEvents(void)
{
  HandleEvents();
  aioPoll(0);
  return 0;
}


int ioMousePoint(void)
{
  ioProcessEvents();  /* process all pending events */
  /* x is high 16 bits; y is low 16 bits */
  return (mousePositionX << 16) | (mousePositionY);
}


/* returns the depth of the Squeak window */
int ioScreenDepth(void)
{
  return qDepth;
}


/* returns the size of the Squeak window */
int ioScreenSize(void)
{
  return ((qWidth << 16) | qHeight);
}

#ifdef REDUNDANT
time_t convertToSqueakTime(time_t unixTime)
{
#ifdef HAVE_TM_GMTOFF
  unixTime+= localtime(&unixTime)->tm_gmtoff;
#else
# ifdef HAVE_TIMEZONE
  unixTime+= ((daylight) * 60*60) - timezone;
# else
#  error: cannot determine timezone correction
# endif
#endif
  /* Squeak epoch is Jan 1, 1901.  Unix epoch is Jan 1, 1970: 17 leap years
     and 52 non-leap years later than Squeak. */
  return unixTime + ((52*365UL + 17*366UL) * 24*60*60UL);
}
#endif /*REDUNDANT*/


/* returns the local wall clock time */
int ioSeconds(void)
{
  return convertToSqueakTime(time(0));
}

int ioSetCursorWithMask(int cursorBitsIndex, int cursorMaskIndex,
			int offsetX, int offsetY)
{
  return 1;
}


int ioSetCursor(int cursorBitsIndex, int offsetX, int offsetY)
{
  ioSetCursorWithMask(cursorBitsIndex, 0, offsetX, offsetY);
  return 0;
}

int ioSetFullScreen(int full)
{
  if (fullScreenFlag == full) {
    return 1;
  }

  fullScreenFlag = full;
  squeak->SetUpWindow();
  redrawFullDisplay();
  return 1;
}


#define bytesPerLine(width, depth)	((((width)*(depth) + 31) >> 5) << 2)
#define bytesPerLineRD(width, depth)	((((width)*(depth)) >> 5) << 2)


int ioForceDisplayUpdate(void)
{
  return 1;
}


#ifdef NON_ZAURUS_ENVIRONMENT
int
ioShowDisplay(int dispBitsIndex, int width, int height, int depth,
		  int affectedL, int affectedR, int affectedT, int affectedB)
{
  if (stDisplayImage == NULL)
    return 0;

  if(affectedR <= affectedL || affectedT >= affectedB)
    return 1;

  if (skipShowDisplay) {
    return 1;
  }

  /* this can happen after resizing the window */
  if (affectedR > width) affectedR= width;
  if (affectedB > height) affectedB= height;

  /*  This happens, too */
  if (affectedR >= qWidth) affectedR= qWidth - 1;
  if (affectedB >= qHeight) affectedB= qHeight - 1;
  
  if ((affectedR <= affectedL) || (affectedT >= affectedB))
    return 1;

  if (depth != qDepth) {
    setDisplayImageFunction(depth, true);
  }

  (*displayImage)((int *)dispBitsIndex, (int*)stDisplayImage->bits(),
		        width, height,
		        affectedL, affectedT, affectedR, affectedB);

  QPainter p(squeak);
  p.drawImage(affectedL, affectedT, *stDisplayImage, affectedL,
	      affectedT, (affectedR - affectedL), (affectedB - affectedT), 0);
}

#else

int
ioShowDisplay(int dispBitsIndex, int width, int height, int depth,
		  int affectedL, int affectedR, int affectedT, int affectedB)
{
  if(affectedR <= affectedL || affectedT >= affectedB)
    return 1;

  if (skipShowDisplay) {
    return 1;
  }
  /* this can happen after resizing the window */
  if (affectedR > width) affectedR= width;
  if (affectedB > height) affectedB= height;

  /*  This happens, too */
  if (affectedR >= qWidth) affectedR= qWidth - 1;
  if (affectedB >= qHeight) affectedB= qHeight - 1;
  
  if ((affectedR <= affectedL) || (affectedT >= affectedB))
    return 1;

  if (depth != qDepth) {
    setDisplayImageFunction(depth, true);
  }

  QDirectPainter p(squeak);
  directp = &p;

  (*displayImage)((int *)dispBitsIndex, (int*)p.frameBuffer(),
		        width, height,
		        affectedL, affectedT, affectedR, affectedB);
  directp = NULL;
  return 1;
}

#endif

int
setDisplayImageFunction(int depth, int bigEndianFlag)
{
  if (depth == 1) {
    displayImage = displayImage1To16;
    qDepth = 1;
    return 1;
  } else if (depth == 8) {
    displayImage = displayImage8To16;
    qDepth = 8;
    return 1;
  } else if (depth == 16) {
    displayImage = displayImage16To16;
    qDepth = 16;
    return 1;
  } else if (depth == 32) {
    displayImage = displayImage32To16;
    qDepth = 32;
    return 1;
  }
  return 0;
}


int ioHasDisplayDepth(int i)
{
  return i == 1 || i == 8 || i == 16 || i == 32;
}


int ioSetDisplayMode(int width, int height, int depth, int fullscreenFlag)
{
  return setDisplayImageFunction(depth, true);
}


void
displayImage1To16(int *dispBitsIndex, int *destBitsIndex,
		    int width, int height,
		    int affectedL, int affectedT, int affectedR, int affectedB)
{
  int scanLine1 = bytesPerLine(width, 1);

  int firstWord1 = scanLine1 * affectedT + bytesPerLineRD(affectedL, 1);
  int lastWord1InLoop = scanLine1 * affectedT + bytesPerLineRD(affectedR, 1);
  int line;

  int aL = (affectedL>>5)<<5;

#ifdef NON_ZAURUS_ENVIRONMENT
  int deltaX = 1;
  int deltaY = stDisplayImage->scanLine(1) - stDisplayImage->scanLine(0);
  int xOffset = 0;
  int yOffset = 0;
#else
  int deltaX;
  int deltaY;
  int xOffset = directp->xOffset();
  int yOffset = directp->yOffset();
#  ifdef IPAQ
  if (portlait) {
    deltaX = - directp->lineStep() / 2;
    deltaY = 2;
  } else {
    deltaX = - 1;
    deltaY = - directp->lineStep();
  }
#  else
  if (portlait) {
    deltaX = 1;
    deltaY = directp->lineStep();
  } else {
    deltaX = directp->lineStep() / 2;
    deltaY = -2;
  }
#  endif
#endif

  unsigned char *vramTopPtr = (unsigned char*)destBitsIndex;

  int vramIndex;
#ifdef NON_ZAURUS_ENVIRONMENT
  vramIndex = (affectedT + yOffset) * deltaY + (aL +xOffset) * 2;
#else
#  ifdef IPAQ
  if (portlait) {
    vramIndex = (239 - (aL + xOffset)) * abs(deltaX) * 2 + (affectedT + yOffset) * abs(deltaY);
  } else {
    vramIndex = (width - 1 - aL + yOffset) * abs(deltaX) * 2 + (239 - (affectedT + xOffset)) * abs(deltaY);
  }
#  else
  if (portlait) {
    vramIndex = (aL + xOffset) * deltaX * 2 + (yOffset + affectedT) * deltaY;
  } else {
    vramIndex = (aL + xOffset) * deltaX * 2 + (((deltaX - 1) - yOffset - affectedT) * 2);
  }
#  endif
#endif

  int maxLine = affectedB;

  for (line = affectedT; line < maxLine; line++) {
    unsigned int *from = (unsigned int*)(((char*)dispBitsIndex) + firstWord1);
    unsigned int *limit;
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);
    int toIndex = aL;

    int m;
    unsigned int mask;
    unsigned int swapedFrom;
    mask = 0x80000000;
    swapedFrom = *from;
    for (m = 0; m < 32; m++) {
      if (toIndex >= affectedL && toIndex < affectedR) {
        *to = (mask & swapedFrom) ? 0 : 0xFFFF;
      }
      mask >>= 1;
      to += deltaX;
      toIndex++;
    }
    from++;

    limit = (unsigned int*)(((char*)dispBitsIndex) + lastWord1InLoop);
    while (from < limit) {
      mask = 0x80000000;
      swapedFrom = *from;
      for (m = 0; m < 32; m++) {
	*to = (mask & swapedFrom) ? 0 : 0xFFFF;
	mask >>= 1;
	to += deltaX;
      }
      toIndex += 32;
      from++;
    }
    mask = 0x80000000;
    swapedFrom = *from;
    for (m = 0; m < 32; m++) {
      if (toIndex >= affectedR) {
        break;
      }
      *to = (mask & swapedFrom) ? 0 : 0xFFFF;
      mask >>= 1;
      to += deltaX;
      toIndex++;
    }
    
    firstWord1 += scanLine1;
    lastWord1InLoop += scanLine1;
    vramIndex += deltaY;
  }

#ifdef USE_CFXGA
  CFXGAShowDisplay2(NULL, yOffset, width, height, 16,
		      affectedL, affectedR, affectedT, affectedB);
#endif
}


void displayImage8To16(int *dispBitsIndex, int *destBitsIndex,
		    int width, int height,
		    int affectedL, int affectedT, int affectedR, int affectedB)
{
  int scanLine8 = bytesPerLine(width, 8);

  int firstWord8 = scanLine8 * affectedT + bytesPerLineRD(affectedL, 8);
  int lastWord8InLoop = scanLine8 * affectedT + bytesPerLineRD(affectedR, 8);
  int line;

  int aL = (affectedL>>2)<<2;

#ifdef NON_ZAURUS_ENVIRONMENT
  int deltaX = 1;
  int deltaY = stDisplayImage->scanLine(1) - stDisplayImage->scanLine(0);
  int xOffset = 0;
  int yOffset = 0;
#else
  int deltaX;
  int deltaY;
  int xOffset = directp->xOffset();
  int yOffset = directp->yOffset();
#  ifdef IPAQ
  if (portlait) {
    deltaX = - directp->lineStep() / 2;
    deltaY = 2;
  } else {
    deltaX = - 1;
    deltaY = - directp->lineStep();
  }
#  else
  if (portlait) {
    deltaX = 1;
    deltaY = directp->lineStep();
  } else {
    deltaX = directp->lineStep() / 2;
    deltaY = -2;
  }
#  endif
#endif
  
  unsigned char *vramTopPtr = (unsigned char*)destBitsIndex;

  int vramIndex;
#ifdef NON_ZAURUS_ENVIRONMENT
  vramIndex = (affectedT + yOffset) * deltaY + (aL +xOffset) * 2;
#else
#  ifdef IPAQ
  if (portlait) {
    vramIndex = (239 - (aL + xOffset)) * abs(deltaX) * 2 + (affectedT + yOffset) * abs(deltaY);
  } else {
    vramIndex = (width - 1 - aL + yOffset) * abs(deltaX) * 2 + (239 - (affectedT + xOffset)) * abs(deltaY);
  }
#  else
  if (portlait) {
    vramIndex = (aL + xOffset) * deltaX * 2 + (yOffset + affectedT) * deltaY;
  } else {
    vramIndex = (aL + xOffset) * deltaX * 2 + (((deltaX - 1) - yOffset - affectedT) * 2);
  }
#  endif
#endif

  int maxLine = affectedB;

  for (line = affectedT; line < maxLine; line++) {
    unsigned char *from = ((unsigned char*)dispBitsIndex) + firstWord8;
    unsigned char *limit;
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);

    int toIndex = aL;
    int i;

    for (i = 3; i >= 0; i--) {
      if (toIndex >= affectedL && toIndex < affectedR) {
	*to = stColors16[from[i]];
      }
      to += deltaX;
      toIndex++;
    }
    from += 4;

    limit = ((unsigned char*)dispBitsIndex) + lastWord8InLoop;

    while (from < limit) {
      *to = stColors16[from[3]];
      to += deltaX;
      *to = stColors16[from[2]];
      to += deltaX;
      *to = stColors16[from[1]];
      to += deltaX;
      *to = stColors16[from[0]];
      to += deltaX;
      toIndex += 4;
      from += 4;
    }

    for (i = 3; i >= 0; i--) {
      if (toIndex >= affectedR) {
	break;
      }
      *to = stColors16[from[i]];
      to += deltaX;
      toIndex++;
    }
    
    firstWord8 += scanLine8;
    lastWord8InLoop += scanLine8;
    vramIndex += deltaY;
  }

#ifdef USE_CFXGA
  CFXGAShowDisplay2(NULL, yOffset, width, height, 16,
		    affectedL, affectedR, affectedT, affectedB);
#endif
}


void displayImage16To16(int *dispBitsIndex, int *destBitsIndex,
		    int width, int height,
		    int affectedL, int affectedT, int affectedR, int affectedB)
{
  int scanLine16 = bytesPerLine(width, 16);

  int firstWord16 = scanLine16 * affectedT + bytesPerLineRD(affectedL, 16);
  int lastWord16InLoop = scanLine16 * affectedT + bytesPerLineRD(affectedR, 16);
  int line;

  int aL = (affectedL>>1)<<1;
  int aR = ((affectedR+1)>>1)<<1;
#ifdef NON_ZAURUS_ENVIRONMENT
  int deltaX = 1;
  int deltaY = stDisplayImage->scanLine(1) - stDisplayImage->scanLine(0);
  int xOffset = 0;
  int yOffset = 0;
#else
  int deltaX;
  int deltaY;
  int xOffset = directp->xOffset();
  int yOffset = directp->yOffset();
#  ifdef IPAQ
  if (portlait) {
    deltaX = - directp->lineStep() / 2;
    deltaY = 2;
  } else {
    deltaX = - 1;
    deltaY = - directp->lineStep();
  }
#  else
  if (portlait) {
    deltaX = 1;
    deltaY = directp->lineStep();
  } else {
    deltaX = directp->lineStep() / 2;
    deltaY = -2;
  }
#  endif
#endif
  
  unsigned char *vramTopPtr = (unsigned char*)destBitsIndex;

  int vramIndex;

#ifdef NON_ZAURUS_ENVIRONMENT
  vramIndex = (affectedT + yOffset) * deltaY + (aL +xOffset) * 2;
#else
#  ifdef IPAQ
  if (portlait) {
    vramIndex = (239 - (aL + xOffset)) * abs(deltaX) * 2 + (affectedT + yOffset) * abs(deltaY);
  } else {
    vramIndex = (width - 1 - aL + yOffset) * abs(deltaX) * 2 + (239 - (affectedT + xOffset)) * abs(deltaY);
  }
#  else
  if (portlait) {
    vramIndex = (aL + xOffset) * deltaX * 2 + (yOffset + affectedT) * deltaY;
  } else {
    vramIndex = (aL + xOffset) * deltaX * 2 + (((deltaX - 1) - yOffset - affectedT) * 2);
  }
#  endif
#endif

  int maxLine = affectedB;

  for (line = affectedT; line < maxLine; line++) {
    unsigned short *from = (unsigned short *)(((unsigned char*)dispBitsIndex) + firstWord16);
    unsigned short *limit;
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);
    unsigned short pixel;

    if (aL != affectedL) {
      to += deltaX;
      pixel = *(from);
      pixel = (((pixel&0x7FE0)<<1) |
	          ((pixel&0x1f)));
      *to = pixel;
      to += deltaX;
      from += 2;
    }

    limit = (unsigned short *)(((unsigned char*)dispBitsIndex) + lastWord16InLoop);

    while (from < limit) {
      pixel = *(from+1);
      pixel = (((pixel&0x7FE0)<<1) |
	          ((pixel&0x1f)));
      *to = pixel;
      to += deltaX;
      pixel = *(from);
      pixel = (((pixel&0x7FE0)<<1) |
	          ((pixel&0x1f)));
      *to = pixel;
      to += deltaX;
      from += 2;
    }

    if (aR != affectedR) {
      pixel = *(from+1);
      pixel = (((pixel&0x7FE0)<<1) |
	     ((pixel&0x1f)));
      *to = pixel;
    }

    firstWord16 += scanLine16;
    lastWord16InLoop += scanLine16;
    vramIndex += deltaY;
  }

#ifdef USE_CFXGA
  CFXGAShowDisplay2(NULL, yOffset, width, height, 16,
		    affectedL, affectedR, affectedT, affectedB);
#endif
}

#if 0
void displayImage16To16CFXGA(int *dispBitsIndex, int *destBitsIndex,
		    int width, int height,
		    int affectedL, int affectedT, int affectedR, int affectedB)
{
  int scanLine16 = bytesPerLine(width, 16);

  int firstWord16 = scanLine16 * affectedT + bytesPerLineRD(affectedL, 16);
  int lastWord16InLoop = scanLine16 * affectedT + bytesPerLineRD(affectedR, 16);
  int line;

  int aL = (affectedL>>1)<<1;
  int aR = ((affectedR+1)>>1)<<1;
#ifdef NON_ZAURUS_ENVIRONMENT
  int deltaX = 1;
  int deltaY = stDisplayImage->scanLine(1) - stDisplayImage->scanLine(0);
  int xOffset = 0;
  int yOffset = 0;
#else
  int deltaX;
  int deltaY;
  int xOffset = directp->xOffset();
  int yOffset = directp->yOffset();
  int pDeltaX;
  int pDeltaY;
  if (portlait) {
    deltaX = 1;
    deltaY = directp->lineStep();
  } else {
    deltaX = directp->lineStep() / 2;
    deltaY = -2;
    pDeltaX = 1;
    pDeltaY = 640*2;
  }
#endif
  
  unsigned char *vramTopPtr = (unsigned char*)destBitsIndex;

  int vramIndex;
  int firstVramIndex;
  int pIndex;
#ifdef NON_ZAURUS_ENVIRONMENT
  vramIndex = (affectedT + yOffset) * deltaY + (aL +xOffset) * 2;
#else
  if (portlait) {
    vramIndex = (aL + xOffset) * deltaX * 2 + (yOffset + affectedT) * deltaY;
  } else {
    vramIndex = (aL + xOffset) * deltaX * 2 + (((deltaX - 1) - yOffset - affectedT) * 2);
    firstVramIndex = vramIndex;
    pIndex = aL * pDeltaX * 2 + affectedT * pDeltaY;
  }
  //int vramIndex = (aL + xOffset) * deltaX * 2 + ((319 - yOffset -  affectedT) * 2);
#endif

  int maxLine = affectedB;

  for (line = affectedT; line < maxLine; line++) {
    unsigned short *from = (unsigned short *)(((unsigned char*)dispBitsIndex) + firstWord16);
    unsigned short *limit;
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);
    unsigned short *to2 = (unsigned short *)(projectorBuffer + pIndex);
    unsigned short pixel;

    if (aL != affectedL) {
      pixel = *(from);
      pixel = (((pixel&0x7FE0)<<1) |
	          ((pixel&0x1f)));
      to += deltaX;
      to2 += pDeltaX;
      *to = pixel;
      *to2 = pixel;
      to += deltaX;
      to2 += pDeltaX;
      from += 2;
    }

    limit = (unsigned short *)(((unsigned char*)dispBitsIndex) + lastWord16InLoop);

    while (from < limit) {
      pixel = *(from+1);
      pixel = (((pixel&0x7FE0)<<1) |
	          ((pixel&0x1f)));
      *to = pixel;
      *to2 = pixel;
      to += deltaX;
      to2 += pDeltaX;
      pixel = *(from);
      pixel = (((pixel&0x7FE0)<<1) |
	          ((pixel&0x1f)));
      *to = pixel;
      *to2 = pixel;
      to += deltaX;
      to2 += pDeltaX;
      from += 2;
    }

    if (aR != affectedR) {
      pixel = *(from+1);
      pixel = (((pixel&0x7FE0)<<1) |
	     ((pixel&0x1f)));
      *to = pixel;
      *to2 = pixel;
    }

    firstWord16 += scanLine16;
    lastWord16InLoop += scanLine16;
    vramIndex += deltaY;
    pIndex += pDeltaY;
  }

#ifdef USE_CFXGA
  CFXGAShowDisplay2(NULL, yOffset, width, height, 16,
		    aL, aR, affectedT, affectedB);
#endif
}
#endif /* 0 */

void
displayImage32To16(int *dispBitsIndex, int *destBitsIndex,
		    int width, int height,
		    int affectedL, int affectedT, int affectedR, int affectedB)
{
  int scanLine32 = bytesPerLine(width, 32);

  int firstWord32 = scanLine32 * affectedT + bytesPerLineRD(affectedL, 32);
  int lastWord32 = scanLine32 * affectedT + bytesPerLine(affectedR, 32);
  int line;

  int aL = affectedL;
  
#ifdef NON_ZAURUS_ENVIRONMENT
  int deltaX = 1;
  int deltaY = stDisplayImage->scanLine(1) - stDisplayImage->scanLine(0);
  int xOffset = 0;
  int yOffset = 0;
#else
  int deltaX;
  int deltaY;
  int xOffset = directp->xOffset();
  int yOffset = directp->yOffset();
#  ifdef IPAQ
  if (portlait) {
    deltaX = - directp->lineStep() / 2;
    deltaY = 2;
  } else {
    deltaX = - 1;
    deltaY = - directp->lineStep();
  }
#  else
  if (portlait) {
    deltaX = 1;
    deltaY = directp->lineStep();
  } else {
    deltaX = directp->lineStep() / 2;
    deltaY = -2;
  }
#  endif
#endif

  unsigned char *vramTopPtr = (unsigned char*)destBitsIndex;

  int vramIndex;
#ifdef NON_ZAURUS_ENVIRONMENT
  vramIndex = (affectedT + yOffset) * deltaY + (aL +xOffset) * 2;
#else
#  ifdef IPAQ
  if (portlait) {
    vramIndex = (239 - (aL + xOffset)) * abs(deltaX) * 2 + (affectedT + yOffset) * abs(deltaY);
  } else {
    vramIndex = (width - 1 - aL + yOffset) * abs(deltaX) * 2 + (239 - (affectedT + xOffset)) * abs(deltaY);
  }
#  else
  if (portlait) {
    vramIndex = (aL + xOffset) * deltaX * 2 + (yOffset + affectedT) * deltaY;
  } else {
    vramIndex = (aL + xOffset) * deltaX * 2 + (((deltaX - 1) - yOffset - affectedT) * 2);
  }
#  endif
#endif

  int maxLine = affectedB;

  for (line = affectedT; line < maxLine; line++) {
    unsigned int *from = (unsigned int *)(((unsigned char*)dispBitsIndex) + firstWord32);
    unsigned int *limit = (unsigned int *)(((unsigned char*)dispBitsIndex) + lastWord32);
    unsigned short *to = (unsigned short *)(vramTopPtr + vramIndex);
    unsigned int pixel;
    while (from < limit) {
      /*
	0000 0000 1111 1111 0000 0000 0000 0000
	                    1111 1000 0000 0000
        0000 0000 0000 0000 1111 1111 0000 0000
	                    0000 0111 1110 0000
	0000 0000 0000 0000 0000 0000 1111 1111
	                    0000 0000 0001 1111
    */
      pixel = *from;
      pixel =
	((pixel&0x00F80000)>>8)
	|((pixel&0x0000FC00)>>5)
	|((pixel&0x000000F8)>>3);
      *to = pixel;
      to += deltaX;
      from++;
    }
    firstWord32 += scanLine32;
    lastWord32 += scanLine32;
    vramIndex += deltaY;
  }
#ifdef USE_CFXGA
  CFXGAShowDisplay2(NULL, yOffset, width, height, 16,
		    affectedL, affectedR, affectedT, affectedB);
#endif
}


/*** power management ***/


int ioDisablePowerManager(int disableIfNonZero)
{
  return true;
}


/*** VM & Image File Naming ***/


/* copy src filename to target, if src is not an absolute filename,
 * prepend the cwd to make target absolute */
void pathCopyAbs(char *target, const char *src, size_t targetSize)
{
  if (src[0] == '/')
  {
    strcpy(target, src);
  }
  else
  {
    getcwd(target, targetSize);
    strcat(target, "/");
    strcat(target, src);
  }
}

void RecordFullPathForVmName(char *localVmName)
{
  /* get canonical path to vm */
  if (realpath(localVmName, vmPath) == 0)
    pathCopyAbs(vmPath, localVmName, sizeof(vmPath));

  /* truncate vmPath to dirname */
  {
    int i= 0;
    for (i= strlen(vmPath); i >= 0; i--)
      if ('/' == vmPath[i])
	{
	  vmPath[i+1]= '\0';
	  return;
	}
  }
}

void RecordFullPathForImageName(char *localImageName)
{
  struct stat s;
  /* get canonical path to image */
  if ((stat(localImageName, &s) == -1) || (realpath(localImageName, imageName) == 0))
    pathCopyAbs(imageName, localImageName, sizeof(255));
}

int imageNameSize(void)
{
  return strlen(imageName);
}

int imageNameGetLength(int sqImageNameIndex, int length)
{
  char *sqImageName= (char *)sqImageNameIndex;
  int count, i;

  count= strlen(imageName);
  count= (length < count) ? length : count;

  /* copy the file name into the Squeak string */
  for (i= 0; i < count; i++)
    sqImageName[i]= imageName[i];

  return count;
}

int imageNamePutLength(int sqImageNameIndex, int length)
{
  char *sqImageName= (char *)sqImageNameIndex;
  int count, i;

  count= (IMAGE_NAME_SIZE < length) ? IMAGE_NAME_SIZE : length;

  /* copy the file name into a null-terminated C string */
  for (i= 0; i < count; i++) imageName[i]= sqImageName[i];
  imageName[count]= 0;

  return count;
}


/*** Timer support ***/


void sigalrm(int signum)
{
  lowResMSecs+= LOW_RES_TICK_MSECS;
}


void SetUpTimers(void)
{
  /* set up the micro/millisecond clock */
  gettimeofday(&startUpTime, 0);
  if (useItimer)
    {
      /* set up the low-res (50th second) millisecond clock */
      /* WARNING: all system calls must check for EINTR!!! */
      {
	struct sigaction sa;
	sigset_t ss1, ss2;
	sigemptyset(&ss1);
	sigprocmask(SIG_BLOCK, &ss1, &ss2);
	sa.sa_handler= sigalrm;
	sa.sa_mask= ss2;
#      ifdef SA_RESTART	/* we're probably on Linux */
	sa.sa_flags= SA_RESTART;
#      else
	sa.sa_flags= 0;	/* assume we already have BSD behaviour */
#      endif
#      ifdef __linux__
	sa.sa_restorer= 0;
#      endif
	sigaction(SIGALRM, &sa, 0);
      }
      {
	struct itimerval iv;
	iv.it_interval.tv_sec= 0;
	iv.it_interval.tv_usec= LOW_RES_TICK_MSECS * 1000;
	iv.it_value= iv.it_interval;
	setitimer(ITIMER_REAL, &iv, 0);
      }
    }
}



/*** X selection handling ***/



void SetUpClipboard(void)
{
  stPrimarySelection= 0;
  stPrimarySelectionSize= 0;
  stOwnsSelection= 0;
}

int clipboardSize(void)
{
  if (stOwnsSelection)
    return stPrimarySelection ? strlen(stPrimarySelection) : 0;
  return strlen(getSelection());
}

static void allocateSelectionBuffer(int count)
{
}

/* claim ownership of the X selection, providing the given string to requestors */
int clipboardWriteFromAt(int count, int byteArrayIndex, int startIndex)
{
  return 0;
}

/* transfer the X selection into the given byte array; optimise local requests */
int clipboardReadIntoAt(int count, int byteArrayIndex, int startIndex)
{
  return 0;
}


/*** Profiling ***/


int clearProfile(void) { return 0; }
int dumpProfile(void) { return 0; }
int startProfiling(void) { return 0; }
int stopProfiling(void) { return 0; }


/*** Access to system attributes and command-line arguments ***/


static char *getAttribute(int id)
{
  if (id < 0) {
    /* VM argument */
    if (-id  < vmArgCnt)
      return vmArgVec[-id];
  } else switch (id) {
    case 0:
      return vmName[0] ? vmName : vmArgVec[0];
    case 1:
      return imageName;
    case 1001:
      /* basic OS type: "unix", "win32", "mac", ... */
      return OS_TYPE;
    case 1002:
      /* specific OS type: "solaris2.5" on unix, "win95" on win32 */
      return VM_HOST_OS;
    case 1003:
      /* processor architecture: "68k", "x86", "PowerPC", ...  */
      return VM_HOST_CPU;
    case 1004:
      return  (char *)interpreterVersion;
    default:
      if ((id-2) < squeakArgCnt)
	return squeakArgVec[id-2];
  }
  success(false);
  return "";
}

int attributeSize(int id)
{
  return strlen(getAttribute(id));
}

int getAttributeIntoLength(int id, int byteArrayIndex, int length)
{
  if (length > 0)
    strncpy((char *)byteArrayIndex, getAttribute(id), length);
  return 0;
}



/*** Command line ***/



static char *progName= 0;

void usage()
{
  printf("Usage: %s [<option>...] [<imageName> [<argument>...]]\n", progName);
  printf("       %s [<option>...] -- [<argument>...]\n\n", progName);
  printf("<option> can be:\n");
  printf("  -help                 print this help message, then exit\n");
  printf("  -jit                  enable the dynamic compiler (if available)\n");
  printf("  -memory <size>[mk]    set initial memory size (default: %dm)\n",
	 DefaultHeapSize);
  printf("  -nomixer              disable modification of sound mixer settings\n");
  printf("  -notimer              disable interval timer for low-res clock \n");
  printf("  -plugins <path>       specify alternative plugin location (see manpage)\n");
  printf("  -version              print version information, then exit\n");
#ifndef HEADLESS
  printf("X11 <option>s:\n");
  printf("  -browserWindow <wid>  run in window <wid>\n");
  printf("  -browserPipes <r> <w> run as Browser plugin using descriptors <r> <w>\n");
  printf("  -cmdmod <n>           map Mod<n> to the Command key\n");
  printf("  -display <dpy>        display on <dpy> (default: $DISPLAY)\n");
  printf("  -fullscreen           occupy the entire screen\n");
  printf("  -headless             run in headless (no window) mode\n");
  printf("  -iconic               start up iconified\n");
  printf("  -lazy                 go to sleep when main window unmapped\n");
  printf("  -mapdelbs             map Delete key onto Backspace\n");
  printf("  -noevents             disable event-driven input support\n");
  printf("  -notitle              disable the Squeak window title bar\n");
  printf("  -optmod <n>           map Mod<n> to the Option key\n");
  printf("  -swapbtn              swap yellow (middle) and blue (right) buttons\n");
  printf("  -xasync               don't serialize display updates\n");
  printf("  -xshm                 use X shared memory extension\n");
#endif
  printf("QPE <option>s:\n");
  printf("  -fullscreen           occupy the entire screen\n");
  printf("  -landscape            show in landscape mode\n");
  if (useJit)
    {
      printf("jit <option>s:\n");
      printf("  -align <n>            align functions at <n>-byte boundaries\n");
      printf("  -jit<o>[,<d>...]      set optimisation [and debug] levels\n");
      printf("  -maxpic <n>           set maximum PIC size to <n> entries\n");
      printf("  -procs <n>            allow <n> concurrent volatile processes\n");
      printf("  -spy                  enable the system spy\n");
    }
  printf("\nNotes:\n");
  printf("  <imageName> defaults to `squeak.image'.\n");
#ifndef HEADLESS
  printf("  Using `unix:0' for <dpy> may improve local display performance.\n");
  printf("  -xshm only works when Squeak is running on the X server host.\n");
#endif
  printf("  <argument>s are ignored, but are processed by the Squeak image.\n");
  printf("  The first <argument> normally names a Squeak `script' to execute.\n");
  printf("  Precede <arguments> by `--' to use default image.\n");
  exit(1);
}


void imageNotFound(char *imageName)
{
  /* image file is not found */
  fprintf(stderr,
	  "Could not open the Squeak image file `%s'.\n"
	  "\n"
	  "There are three ways to open a Squeak image file.  You can:\n"
	  "  1. Put copies of the default image and changes files in this directory.\n"
	  "  2. Put the name of the image file on the command line when you\n"
	  "     run squeak (use the `-help' option for more information).\n"
	  "  3. Set the environment variable SQUEAK_IMAGE to the name of the image\n"
	  "     that you want to use by default.\n"
	  "\n"
	  "For more information, type: `man squeak' (without the quote characters).\n",
	  imageName);
  exit(1);
}


void versionInfo(void)
{
  extern int vm_serial;
  extern char *vm_date, *cc_version, *ux_version;

  printf("%s #%d", VM_VERSION, vm_serial);
#if defined(USE_XSHM)
  printf(" XShm");
#endif
  printf(" %s %s\n", vm_date, cc_version);
  printf("%s\n", interpreterVersion);
  printf("%s\n", ux_version);
  printf("default plugin location: %s/*.so\n", VM_LIBDIR);
  exit(0);
}


void outOfMemory(void)
{
  fprintf(stderr, "out of memory\n");
  exit(1);
}


int strtobkm(const char *str)
{
  char *suffix;
  int value= strtol(str, &suffix, 10);
  switch (*suffix)
    {
    case 'k': case 'K':
      value*= 1024;
      break;
    case 'm': case 'M':
      value*= 1024*1024;
      break;
    }
  return value;
}


static int jitArgs(char *str)
{
  char *endptr= str;
  int  args= 3;				// default JIT mode = fast compiler

  if (*str == '\0') return args;
  if (*str != ',')
    args= strtol(str, &endptr, 10);	// mode
  while (*endptr == ',')		// [,debugFlag]*
    args|= (1 << (strtol(endptr + 1, &endptr, 10) + 8));
  return args;
}


void ParseEnvironment(void)
{
  char *ev= 0;

  if ((ev= getenv("SQUEAK_IMAGE")))	strcpy(shortImageName, ev);
  //else					strcpy(shortImageName, "squeak.image");
  if ((ev= getenv("SQUEAK_MEMORY")))	initialHeapSize= strtobkm(ev);
  if ((ev= getenv("SQUEAK_PLUGINS")))	squeakPlugins= strdup(ev);
  if ((ev= getenv("SQUEAK_NOMIXER")))	noSoundMixer= 1;
  if ((ev= getenv("SQUEAK_NOTIMER")))	useItimer= 0;

#ifndef HEADLESS
  if (getenv("SQUEAK_LAZY"))		sleepWhenUnmapped= 1;
  if (getenv("SQUEAK_SPY"))		withSpy= 1;
  if (getenv("SQUEAK_NOTITLE"))		noTitle= 1;
  if (getenv("SQUEAK_FULLSCREEN"))	fullScreen= 1;
  if (getenv("SQUEAK_ICONIC"))		iconified= 1;
  if (getenv("SQUEAK_NOEVENTS"))	noEvents= 1;
  if (getenv("SQUEAK_MAPDELBS"))	mapDelBs= 1;
  if (getenv("SQUEAK_SWAPBTN"))		swapBtn= 1;
  if ((ev= getenv("SQUEAK_OPTMOD")))	optMapIndex= Mod1MapIndex + atoi(ev) - 1;
  if ((ev= getenv("SQUEAK_CMDMOD")))	cmdMapIndex= Mod1MapIndex + atoi(ev) - 1;
#if defined(USE_XSHM)
  if (getenv("SQUEAK_XSHM"))		useXshm= 1;
  if (getenv("SQUEAK_XASYNC"))		asyncUpdate= 1;
#endif
#endif
  if ((ev= getenv("SQUEAK_JIT")))	useJit= jitArgs(ev);
  if ((ev= getenv("SQUEAK_ALIGN")))	asmAlign= atoi(ev);
  if ((ev= getenv("SQUEAK_PROCS")))	jitProcs= atoi(ev);
  if ((ev= getenv("SQUEAK_MAXPIC")))	jitMaxPIC= atoi(ev);
  if (asmAlign <  1) asmAlign= 1;
  if (asmAlign > 16) asmAlign= 16;
}


/* read options from the header of the specified image */

#define HEADER_SIZE 512

static void readHeaderOptions(const char *filename)
{
  int   headerArgc= 0;
  char *headerArgv[512];
  char *header= (char*)malloc(HEADER_SIZE+1);

  if (header == NULL)
    outOfMemory();
  /* read the header */
  {
    int fd= open(filename, O_RDONLY);
    if (fd < 0) return;
    if (read(fd, header, HEADER_SIZE) < HEADER_SIZE)
      {
	free(header);
	close(fd);
	return;
      }
    header[HEADER_SIZE]= '\0';  /* make sure header is a valid C string */
    close(fd);
  }
  if (header[0] != '#' || header[1] != '!')
    {
      /* not a Unix image file */
      free(header);
      return;
    }
  /* find the newline which marks the beginning of the options */
  {
    int headerPos= 0;
    while ((headerPos < HEADER_SIZE) && (header[headerPos] != '\n'))
      headerPos++;
    headerPos++;   /* skip the newline */
    /* read in options, one at a time; we are finished when either the
       header is exhausted, or a NUL character is seen at the beginning
       of an option */
    while ((headerPos < HEADER_SIZE) && (header[headerPos] != '\0'))
      {
	/* record one option */
	headerArgv[headerArgc]= &header[headerPos];
	headerArgc+= 1;
	/* find the begining of the next option */
	while (headerPos < HEADER_SIZE && header[headerPos] != '\0')
	  headerPos+= 1;
	/* skip over the '\0' */
	headerPos+= 1;
      }
  }
  /* got all options -- process them */
  ParseArguments(headerArgc, headerArgv, 1);
}


void ParseArguments(int argc, char *argv[], int inHeader)
{
# define skipArg()	(--argc, argv++)
# define saveArg()	(vmArgVec[vmArgCnt++]= *skipArg())
  int argsInHeader= 0;

  if (vmArgCnt == 0)
    saveArg();		/* vm name */

  while ((argc > 0) && (**argv == '-'))	/* more options to parse */
    {
      char *arg= *argv;
      if (!strcmp(arg, "--"))		/* escape from option processing */
	break;

      saveArg();

      if      (!strcmp(arg, "-help"))		usage();
      if      (!strcmp(arg, "-nomixer"))	noSoundMixer= 1;
      if      (!strcmp(arg, "-notimer"))	useItimer= 0;
#    ifndef HEADLESS
      else if (!strcmp(arg, "-headless"))	headless= 1;
#      if defined(USE_XSHM)
      else if (!strcmp(arg, "-xshm"))		useXshm= 1;
      else if (!strcmp(arg, "-xasync"))		asyncUpdate= 1;
#      endif
      else if (!strcmp(arg, "-lazy"))		sleepWhenUnmapped= 1;
      else if (!strcmp(arg, "-noevents"))	noEvents= 1;
      else if (!strcmp(arg, "-notitle"))	noTitle= 1;
      else if (!strcmp(arg, "-mapdelbs"))	mapDelBs= 1;
      else if (!strcmp(arg, "-swapbtn"))	swapBtn= 1;
#    endif
      else if (!strncmp(arg, "-jit", 4))	useJit= jitArgs(arg+4);
      else if (!strcmp(arg, "-nojit"))		useJit= 0;
      else if (!strcmp(arg, "-spy"))		withSpy= 1;
      else if (!strcmp(arg, "-fullscreen"))	fullScreen= 1;
#    ifndef HEADLESS
      else if (!strcmp(arg, "-iconic"))		iconified= 1;
#    endif
      else if (!strcmp(arg, "-version"))	versionInfo();
      else if (!strcmp(arg, "-args_in_header"))
	{
	  argsInHeader= 1;
	  if (argc > 0)
	    /* record the image name */
	    strcpy(shortImageName, *skipArg());
	  /* read the arguments */
	  if (!inHeader)   /* but don't recurse */
	    readHeaderOptions(shortImageName);
	}
      else
	{
	  /* option requires an argument */
	  if      (arg == 0)			usage();
	  else if (!strcmp(arg, "-align"))
	    {
	      asmAlign= atoi(saveArg());
	      if (asmAlign <  1) asmAlign= 1;
	      if (asmAlign > 16) asmAlign= 16;
	    }
	  else if (!strcmp(arg, "-procs"))	jitProcs= atoi(saveArg());
	  else if (!strcmp(arg, "-maxpic"))	jitMaxPIC= atoi(saveArg());
	  else if (!strcmp(arg, "-memory"))	initialHeapSize= strtobkm(saveArg());
	  else if (!strcmp(arg, "-plugins"))	squeakPlugins= strdup(saveArg());
#        if !defined(HEADLESS)
	  else if (!strcmp(arg, "-display"))	displayName= saveArg();
	  else if (!strcmp(arg, "-optmod"))
	    optMapIndex= Mod1MapIndex + atoi(saveArg()) - 1;
	  else if (!strcmp(arg, "-cmdmod"))
	    cmdMapIndex= Mod1MapIndex + atoi(saveArg()) - 1;
	  else if (!strcmp(arg, "-browserWindow"))
	    {
	      sscanf(saveArg(), "%li", &browserWindow);
	      if (browserWindow == 0)
		{
		  fprintf(stderr, "Error: invalid -browserWindow argument!\n");
		  exit(1);
		}
	    }
	  else if (!strcmp(arg, "-browserPipes"))
	    {
	      if (argc < 2) usage();
	      sscanf(saveArg(), "%i", &browserPipes[0]);
	      sscanf(saveArg(), "%i", &browserPipes[1]);
	      /* receive browserWindow */
#	     ifdef DEBUG_BROWSER
	      fprintf(stderr, "browser: reading window\n");
#	     endif
	      if (sizeof(browserWindow) !=
		   read(browserPipes[0], (void *)&browserWindow, sizeof(browserWindow)))
		{
		  perror("reading browserWindow");
		  exit(1);
		}
#	     ifdef DEBUG_BROWSER
	      fprintf(stderr, "browser: window = 0x%x\n", browserWindow);
#	     endif
	    }
#        endif
	  else if (!strcmp(arg, "-landscape")) portlait = false;
	  else
	    usage();
	} /* option with argument */
    } /* while(more options) */

  if (argc > 0)
    {
      if (!strcmp(*argv, "--"))
	skipArg();
      else
	{
	  /* look for an image name */
	  if (!argsInHeader)
	    {  /* if args_in_header, then the image name will have already been read  */
	      if (inHeader)
		skipArg();
	      else
		{
		  strcpy(shortImageName, saveArg());
		  if (0 == strstr(shortImageName, ".image"))
		    strcat(shortImageName, ".image");
		}
	    }
	}
    }
  /* save remaining arguments as Squeak arguments */
  while (argc > 0)
    squeakArgVec[squeakArgCnt++]= *skipArg();

# undef saveArg
# undef skipArg
}



/*** internal image ***/



#if defined(USE_INTERNAL_IMAGE)

  /*-- EXPERIMENTAL -- read [and inflate] an internal image file --*/

  int sqImageFileClose(FILE *f)
  {
    int err= 0;
    if (f != 0)
      return fclose(f);
    assert(internalImage != 0);
    if (internalZStream != 0)
      {
	printf("decompressed %ld bytes\n", ztell(internalZStream));
	err= zclose(internalZStream);
      }
    internalImage= 0;
    internalImagePtr= 0;
    internalZStream= 0;
    return err;
  }

  int sqImageFilePosition(FILE *f)
  {
    if (f != 0)
      return ftell(f);
    assert(internalImage != 0);
    if (internalZStream != 0)
      return ztell(internalZStream);
    return internalImagePtr - internalImage;
  }

  int sqImageFileRead(void *ptr, size_t sz, size_t count, FILE *f)
  {
    if (f != 0)
      return fread(ptr, sz, count, f);
    assert(internalImage != 0);
    if (internalZStream != 0)
      return zread(ptr, sz, count, internalZStream);
    memcpy(ptr, (void *)internalImagePtr, sz * count);
    internalImagePtr+= sz*count;
    return sz*count;
  }

  int sqImageFileSeek(FILE *f, long pos)
  {
    if (f != 0)
      return fseek(f, pos, SEEK_SET);
    assert(internalImage != 0);
    if (internalZStream != 0)
      return zseek(internalZStream, pos, SEEK_SET);
    internalImagePtr= internalImage + pos;
    return 0;
  }

/* get a value for RTLD_NOW, with increasing levels of desperation... */

#if !defined(RTLD_NOW)
# if defined(DL_NOW)
#   define RTLD_NOW DL_NOW
# elif defined(RTLD_LAZY)
#   define RTLD_NOW RTLD_LAZY
# elif defined(DL_LAZY)
#   define RTLD_NOW DL_LAZY
# else
#   warning: defining RTLD_NOW as 1
#   define RTLD_NOW 1
# endif
#endif

  int openInternal(void)
  {
    unsigned char *internalImageEnd= 0;
    void *handle= dlopen(0, RTLD_NOW);
    if (handle == 0)
      {
	fprintf(stderr, "dlopen: %s\n", dlerror());
	exit(1);
      }
    /* non-zero means use in-memory file operations */
    internalImage= (unsigned char *)dlsym(handle, "__squeak_image_start");
    if (internalImage != 0)
      {
	internalImageEnd= (unsigned char *)dlsym(handle, "__squeak_image_end");
	internalImagePtr= internalImage;
	printf("reading internal image at 0x%08x + %d\n",
	       (unsigned)internalImage,
	       (int)(internalImageEnd - internalImage));
	strcpy(shortImageName, "internal.image");
      }
    else
      {
	internalImage= (unsigned char *)dlsym(handle, "__squeak_image_gz_start");
	if (internalImage == 0)
	  return 0;
	internalImageEnd= (unsigned char *)dlsym(handle, "__squeak_image_gz_end");
	{
	  char name[64], comment[64];
	  if (0 == gzstat(internalImage,
			  name, sizeof(name), comment, sizeof(comment)))
	    {
	      fprintf(stderr, "internal image: %s\n", z_error);
	      exit(1);
	    }
	  printf("decompressing %s at 0x%08x + %d\n",
		 ((*name == '\0') ? "internal image" : name),
		 (unsigned)internalImage,
		 (int)(internalImageEnd - internalImage));
	  if (*name == '\0')
	    strcpy(shortImageName, "internal.image");
	  else
	    strcpy(shortImageName, name);
	  if (*comment != '\0')
	    printf("%s\n", comment);
	}
	/* non-zero means inflate on-the-fly */
	internalZStream= gzopen(internalImage, internalImageEnd - internalImage);
	if (internalZStream == 0)
	  {
	    fprintf(stderr, "zopen: %s\n", z_error);
	    exit(1);
	  }
      }
    dlclose(handle);
    return internalImage != 0;
  }

#endif /* USE_INTERNAL_IMAGE */


/*** Segmentation fault handler ***/


#include <signal.h>

void sigsegv(int ignore)
{
  error("Segmentation fault");
}

void sighup(int ignore)
{
  printCallStack();
  nextWakeupTick= ioLowResMSecs();
  interruptPending= true;
  interruptCheckCounter= 0;
#if !defined(HEADLESS)
  signalInputEvent();
#endif
}

#ifdef __alpha__
/* headers for setsysinfo (see below) */
# include <sys/sysinfo.h>
# include <sys/proc.h>
#endif


void
SqueakWindow::focusInEvent(QFocusEvent *evt) {
  SetUpWindow();
}

void
SqueakWindow::enableFullScreen()
{
  if (!interpreterStarted) {
    return;
  }

  setFixedSize(stApp->desktop()->size());
  showNormal();
  reparent(0, WStyle_Customize | WStyle_NoBorder, QPoint(0, 0));
  showFullScreen();
}

void
SqueakWindow::disableFullScreen()
{
  if (!interpreterStarted) {
    return;
  }

  //reparent(0, defaultWFlags, defaultOrigin);
  setSizePolicy(defaultSizePolicy);
  setGeometry(defaultGeometry);
  showNormal();
}


QWidget*
SqueakWindow::openFileDialog()
{
  QRect geom = stApp->desktop()->geometry();
#ifdef USE_SL_FILESELECTOR
  fileSelector = new SlFileSelector("/home",
				    "*",
				    this,
				    "Choose a file name");
#else 
  fileSelector = new FileSelector(QString::null,
				  this,
				  "file selector",
				  false,
				  false);
#endif

  fileSelector->resize(geom.width(), geom.height()-32);

#ifdef USE_SL_FILESELECTOR
  connect(fileSelector, SIGNAL(fileSelected(const QFileInfo&)), this, SLOT(setShortImageName(const QFileInfo&)));
#else 
  connect(fileSelector, SIGNAL(fileSelected(const DocLnk&)), this, SLOT(setShortImageName(const DocLnk&)));
#endif
  return fileSelector;
}

/*** ...we came in? ***/

int
main(int argc, char **argv, char **envp)
{
  /* check the interpreter's size assumptions for basic data types */
  if (sizeof(int)    != 4) error("This C platform's integers are not 32 bits.");
  if (sizeof(double) != 8) error("This C platform's floats are not 64 bits.");
  if (sizeof(time_t) != 4) error("This C platform's time_t's are not 32 bits.");

  /* Make parameters global for access from pluggable primitives */
  argCnt= argc;
  argVec= argv;
  envVec= envp;

  /* Allocate arrays to store copies of pointers to command line
     arguments.  Used by getAttributeIntoLength(). */
  if ((vmArgVec= (char**)calloc(argc + 1, sizeof(char *))) == 0)
    outOfMemory();

  if ((squeakArgVec= (char**)calloc(argc + 1, sizeof(char *))) == 0)
    outOfMemory();

  signal(SIGSEGV, sigsegv);

  progName= argv[0];
  shortImageName[0] = 0;

  QPEApplication app(argc, argv);
  stApp = &app;

  initialHeapSize = availableMemory();

  ParseEnvironment();
  ParseArguments(argc, argv, 0);

  fprintf(stderr, "name: %s\n", shortImageName);

  QColor color(128, 128, 128);
  
  squeak = new SqueakWindow();
  if (qstrlen(shortImageName) == 0) {
    squeak->openFileDialog();
  }
  app.showMainWidget(squeak);
  
  squeak->setBackgroundColor(color);
  squeak->SetupQPEDisplay();
  squeak->setCaption("Squeak!");
  if (qstrlen(shortImageName) != 0) {
    squeak->run();
  }
  QPEApplication::grabKeyboard();
  int result = app.exec();
  QPEApplication::ungrabKeyboard();
  return result;
}

void*
startInterpreter(void *arg)
{
  unsigned int available = 0;
  /* initialisation */

  SetUpTimers();

  //sqFileInit();
  //joystickInit();

# if defined(HAVE_TZSET)
  tzset();	/* should _not_ be necessary! */
# endif

  /* if (!realpath(argv[0], vmName))
    vmName[0]= 0; */ /* full VM name */

  /* read the image file and allocate memory for Squeak heap */
  {
    FILE *f= 0;

#   if defined(USE_INTERNAL_IMAGE) /* EXPERIMENTAL */
    if (openInternal())
      {
	RecordFullPathForImageName(shortImageName);
      }
    else
#   endif
      {
	/* strcpy(shortImageName, "/mnt/cf/bench3.image"); */
	if (0 == (f= fopen(shortImageName, "r")))
	  imageNotFound(shortImageName);
	else
	  RecordFullPathForImageName(shortImageName); /* full image path */
      }
    readImageFromFileHeapSize(f, initialHeapSize);
    sqImageFileClose(f);
  }

#if defined(HAVE_LIBDL) && defined(J_ENABLED) && !defined(J_PROFILE)
  {
    /* first try to find an internal dynamic compiler... */
    int handle= ioLoadModule(0);
    int comp= ioFindExternalFunctionIn("j_interpret", handle);
    /* ...and if that fails... */
    if ((comp == 0) && !noJitter)
      {
	/* ...try to find an external one */
	handle= ioLoadModule("SqueakCompiler");
	if (handle != 0)
	  comp= ioFindExternalFunctionIn("j_interpret", handle);
      }
    if (comp)
      {
	((void(*)(void))comp)();
	return 0;
      }
    else
      printf("could not find j_interpret\n");
  }
#endif /*HAVE_LIBDL*/

  squeak->SetUpWindow();

  /* run Squeak */
#ifndef J_PROFILE
  interpret();
#else
  j_interpret();
#endif

  return 0;
}


int ioExit(void)
{
  exit(0);
}

void
SqueakWindow::run()
{
  pthread_attr_t attr;
  pthread_attr_init(&attr);
  if (pthread_create(&interpThread, &attr, startInterpreter, NULL) != 0) {
    fprintf(stderr, "error thread creation\n");
    exit(1);
  }
  interpreterStarted = true;
}

void
SqueakWindow::setShortImageName(const DocLnk& fileInfo)
{
#ifdef USE_SL_FILESELECTOR
  if (fileSelector != NULL) {
    qstrcpy(shortImageName, fileSelector->currentDir());
    strcat(shortImageName, "/");
    strcat(shortImageName, fileInfo.fileName());
  }
    
  if (fileSelector != NULL) {
    fileSelector->hide();
    //delete fileSelector;
  }
  squeak->run();
#endif
}

#ifdef USE_SL_FILESELECTOR

void
SqueakWindow::setShortImageName(const QFileInfo& doc)
{
  strcpy(shortImageName, doc.file());
  if (fileSelector != NULL) {
    fileSelector->hide();
    //delete fileSelector;
  }
  squeak->run();
}

#endif

void
SqueakWindow::closeFileSelector()
{
}

/* meminfo checking */

struct _meminfo {
  unsigned int memTotal;
  unsigned int memFree;
  unsigned int memShared;
  unsigned int buffers;
  unsigned int cached;
  unsigned int swapCached;
  unsigned int active;
  unsigned int inactive;
  unsigned int inact_dirty;
  unsigned int inact_clean;
  unsigned int inact_target;
  unsigned int highTotal;
  unsigned int highFree;
  unsigned int lowTotal;
  unsigned int lowFree;
  unsigned int swapTotal;
  unsigned int swapFree;
  unsigned int commited_as;
} meminfo;

static int
fillMemInfo()
{
  FILE *f = fopen("/proc/meminfo", "r");
  char buffer[128];
  char label[128];
  char number[128];

  if (f == NULL) {
    return false;
  }

  memset(&meminfo, 0, sizeof(meminfo));

  while (fgets(buffer, 128, f)) {
    strcpy(label, strtok(buffer, ": \t"));
    strcpy(number, strtok(NULL, ": \t"));

    if (strcmp(label, "MemFree") == 0) {
      meminfo.memFree = strtobkm(number) * 1024;
      continue;
    }
    if (strcmp(label, "Cached") == 0) {
      meminfo.cached = strtobkm(number) * 1024;
      continue;
    }
    if (strcmp(label, "SwapCached") == 0) {
      meminfo.swapCached = strtobkm(number) * 1024;
      continue;
    }
    if (strcmp(label, "Inactive") == 0) {
      meminfo.inactive = strtobkm(number) * 1024;
      continue;
    }
    if (strcmp(label, "SwapTotal") == 0) {
      meminfo.swapTotal = strtobkm(number) * 1024;
      continue;
    }
    if (strcmp(label, "SwapFree") == 0) {
      meminfo.swapFree = strtobkm(number) * 1024;
      continue;
    }
  }
  return true;
}

unsigned int
availableMemory()
{
  fillMemInfo();

  if (meminfo.swapFree == 0) {
    return (meminfo.cached + meminfo.memFree) * 13 / 20;
  }

  return (meminfo.cached + meminfo.memFree) * 13 / 20 + (5 * 1024 * 1024);
}

#ifdef USE_CFXGA
static bool
isCFXGAAvailable()
{
  FILE* f = fopen("/var/lib/pcmcia/stab", "r");
  if (!f) f = fopen("/var/state/pcmcia/stab", "r");
  if (!f) f = fopen("/var/run/stab", "r");
  if (f) {
    char line[1024];
    while (fgets( line, 1024, f )) {
      if (QString(line).find("cfxga") > 0) {
	if (logger != NULL) {
	  fprintf(logger, "cfxga found\n");
	}
	fclose(f);
	return true;
      }
    }
    fclose(f);
  }
  if (logger != NULL) {
    fprintf(logger, "cfxga not found\n");
  }
  return false;
}

// in case if card is suspended
static bool
CFXGAResumeCard()
{  
  if (!isCFXGAAvailable()){ // check card
    return false;
  }
  system("/sbin/cardctl resume");
  return true;
}

static bool
CFXGAIsActive()
{
  return (cfxgaFd != 0);
}

static void
CFXGASetDefaultSize()
{
  int mode;
  mode = 0;
  ioctl(cfxgaFd, CFXGA_SET_RESO, &mode);
}

static int
CFXGAOpenFixedFB()
{
  int fbHd = 0;
  if ((fbHd = ::open("/dev/fb0", O_RDONLY)) <= 0) {
    return false;
  }

  struct fb_var_screeninfo vinfo;
  if (::ioctl(fbHd, FBIOGET_VSCREENINFO, &vinfo) < 0) {
    close(fbHd);
    return false;
  }
  if (vinfo.xres != 480) {
    close(fbHd);
    return false;
  }

  fix.smem_start = 0 ;
  if (::ioctl(fbHd, FBIOGET_FSCREENINFO, &fix) < 0) {
    close(fbHd);
    return false;
  }

  close(fbHd);
  return true;
}

static bool
CFXGAActivate(bool needToClearScreen) // dispOutInit
{
  int fd;

  if (cfxgaFd != 0) {
    if (logger != NULL) {
      fprintf(logger, "cfxga non zero\n");
    }
    return false;
  }
  //prevOutSize = QSize(QCOORD_MAX, QCOORD_MAX);
  if ((fd = ::open("/dev/cfxga0a", O_RDONLY)) < 0) {
    if (!CFXGAResumeCard()) { // in case if card is suspended
      if (logger != NULL) {
	fprintf(logger, "not resumable\n");
      }
      return false;
    }
    // retry
    for (int i = 0; i < 5; i++){
      if ((fd = ::open("/dev/cfxga0a", O_RDONLY)) >= 0){
	break;
      }
      sleep(1);
    }
    if (fd < 0){
      if (logger != NULL) {
	fprintf(logger, "cannot open\n");
      }
      return false;
    }
  }

  status_t st;
  ::ioctl(fd, CFXGA_GET_STATUS, (char*)&st);
  if ((st.flags & CFXGA_INSTALLED) == 0) {
    if (logger != NULL) {
      fprintf(logger, "status error\n");
    }
    ::close(fd);
    return false;
  }
  //::ioctl(fd, CFXGA_SET_RESO, (char*)&dMode);
  if (needToClearScreen) {
    ::ioctl(fd, CFXGA_CLEAR_SCREEN, (char*)NULL);
  }
  cfxgaFd = fd;
  if (logger != NULL) {
    fprintf(logger, "cfxga: %d\n", cfxgaFd);
  }
  CFXGASetDefaultSize();
  return true;
}

static bool
CFXGAActivate2(bool needToClearScreen) // dispOutInit
{
  if (CFXGAActivate(needToClearScreen)) {
    return CFXGAOpenFixedFB();
  }

  return false;
}

static void
CFXGADeactivate() // dispOutClose
{
  if (cfxgaFd == 0) {
    return;
  }
  ::ioctl(cfxgaFd, CFXGA_CLEAR_SCREEN, (char*)NULL);
  ::close(cfxgaFd);
  cfxgaFd = 0;
}

static QSize
CFXGAOutputSize()
{
  CFXGAActivate(false);
  return QSize(640, 480);
  
//   if(cfxgaFd == 0){
//     if(dMode == CRT_800_600_60_16) {
//       return QSize(800, 600);
//     } else {
//       return QSize(640, 480);
//     }
//   }

//   reso_info_t rt;
//   rt.mode_no = dMode;
//   ::ioctl(cfxgaFd, CFXGA_GET_RESOINFO, (char*)&rt);
//   return QSize(rt.reso.cx, rt.reso.cy);
}

static int
CFXGAShowDisplay(unsigned char* vramPtr, int width, int height, int depth,
		 int affectedL, int affectedR, int affectedT, int affectedB)
{
  copy_info_t ci;

  if (cfxgaFd == 0) {
    return -1;
  }
  ci.x = affectedL;
  ci.y = affectedT;
  ci.w = affectedR - affectedL;
  ci.h = affectedB - affectedT;
  ci.bpp = 16;
  ci.stride = 640*2;
  ci.adr = vramPtr + (affectedT*640*2) + affectedL*2;

  if (logger != NULL) {
    fprintf(logger, "cfxga show: (%d, %d, %d, %d)\n", ci.x, ci.y, ci.w, ci.h);
  }


  return ioctl(cfxgaFd, CFXGA_COPY_FB, &ci);
}

static int
CFXGAShowDisplay2(unsigned char* vramPtr, int offset, int width, int height, int depth,
		 int affectedL, int affectedR, int affectedT, int affectedB)
{
  copy2_info_t ci;

  if (cfxgaFd == 0) {
    return -1;
  }

  if (fix.smem_start == 0) {
    return -1;
  }

  if (portlait) {
    return -1;
  }

  ci.bpp = 16;
  ci.phys_adr = fix.smem_start;
  ci.size = fix.smem_len;
  ci.swivel = SWIVEL_90;
    ci.x = affectedL;
    ci.y = affectedT;
    ci.w = affectedB - affectedT;
    ci.h = affectedR - affectedL;
    ci.stride = 480*2;
    ci.offset = affectedL*480*2+(480-affectedB-offset)*2;

  if (logger != NULL) {
    fprintf(logger, "cfxga show2: (%d, %d, %d, %d)\n", ci.x, ci.y, ci.w, ci.h);
  }

  return ioctl(cfxgaFd, CFXGA_COPY_FB2, &ci);
}

#endif /* USE_CFXGA */
